
/*-
 * maptostackmap_vnodeops.c
 *
 *	@(#)maptostackmap_vnodeops.c	1.2 (UCLA Ficus Project) 2/15/95
 *
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/param.h>
#include <sys/time.h>
#include <sys/vnode.h>
#include <i405/i405.h>
#include <sys/debug.h>   /* for ASSERT */
#include <sys/errno.h>
#include <vm/seg.h>



#if 0
#define DODEBUG(A) A
#else
#define DODEBUG(A)
#endif


/*
 * Set to monitor {get,put}page operations.
 */
static int maptostackmap_bug = 0;



/*-
 *--------------------------------------------------------------------------
 *
 * maptostackmap_getpage -- a generic getpage for the SVCM
 *
 * DESCRIPTION:
 *	First aquire page ownership, then map it in directly with
 *	VOP_NEWGETPAGE.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
maptostackmap_getpage (ap)
	struct nvop_getpage_args *ap;
{
	USES_NVOP_STACKGETPAGE;
	USES_NVOP_GET_SVCM_NAME;
	int error = 0;
	struct vnode *vp = ap->a_vp;
	struct svcm_name *svcm_name;

	DODEBUG(if (maptostackmap_bug)
		printf ("maptostackmap_getpage (%x,%x,%x,%x,%x,%x,%x,%x,%d,%x)\n",
			vp,
			ap->a_off, ap->a_len,
			ap->a_protp,
			ap->a_pl, ap->a_plsz,
			ap->a_seg, ap->a_addr,
			ap->a_rw, ap->a_cred));

	if (error = NVOP_GET_SVCM_NAME(vp, &svcm_name, ap->a_cred))
		return error;

	if (error = svcm_lock(svcm_name->lock))
		return error;

	DODEBUG(pvn_dump_pages(vp));
	error = NVOP_STACKGETPAGE(vp, svcm_name, vp,
			       ap->a_off, ap->a_len,
			       ap->a_protp,
			       ap->a_pl, ap->a_plsz,
			       ap->a_seg, ap->a_addr,
			       ap->a_rw, ap->a_cred);
	DODEBUG(pvn_dump_pages(vp));

	(void)svcm_unlock(svcm_name->lock);
	return error;
}


/*-
 *--------------------------------------------------------------------------
 *
 * maptostackmap_putpage -- map vop_putpage to vop_stackputpage
 *
 * DESCRIPTION:
 *	First aquire page ownership, then map it in directly with
 *	VOP_NEWPUTPAGE.
 *
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
maptostackmap_putpage (ap)
	struct nvop_putpage_args *ap;
{
	USES_NVOP_STACKPUTPAGE;
	USES_NVOP_GET_SVCM_NAME;
	int error;
	struct vnode *vp = ap->a_vp;
	struct svcm_name *svcm_name;

	DODEBUG(if (maptostackmap_bug_bypass)
		printf ("maptostackmap_putpage (%x,%x,%x,%x,%x)\n",
			vp,
			ap->a_off, ap->a_len,
			ap->a_flags,
			ap->a_cred));

	if (error = NVOP_GET_SVCM_NAME(vp, &svcm_name, ap->a_cred))
		return error;

	if (error = svcm_lock(svcm_name->lock))
		return error;

	/*
	 * Normally getpage will fill in svcm_name,
	 * but if the user's writing a new file,
	 * pvn_rdwr may just alloc its own pages, and
	 * so no name may be aquired.
	 *    HMMM...for svcm ownership to work,
	 * we assume getpage will aquire ownership.
	 * In the just-discussed case, it won't.
	 *    The solution is to force pvn_rdwr to pre-allocate pages,
	 * going through getpage and the svcm.  This is also
	 * important because the ufs needs to allocate backing store
	 * at that time.
	 *
	 * The page-out daemon also calls putpage,
	 * perhaps even before getpage has done its thing,
	 * so there are times when we mustwe _have_ to do this here.
	 *
	 * But all real pages _must_ be allocated through
	 * getpage.  Any other approach is an error.
	 * To detect this bug, the newputpage should
	 * check (if we want to be paranoid).
	 */
#if 0
	if (!vp->v_svcm_name)
		if (error = maptostackmap_name_vp(vp, ap->a_cred))
			return errror;
#else
	/*
	 * Panic is above.
	 */
#endif
	DODEBUG(pvn_dump_pages(vp));
	error = NVOP_STACKPUTPAGE(vp, svcm_name, vp,
			       ap->a_off, ap->a_len,
			       ap->a_flags,
			       ap->a_cred);
	DODEBUG(pvn_dump_pages(vp));

	(void)svcm_unlock(svcm_name->lock);
	return error;
}




struct fwl_vnodeopv_entry_desc maptostackmap_fwl_vnodeop_entries[] = {
	{ &nvop_getpage_desc, maptostackmap_getpage, 0, NULL },
	{ &nvop_putpage_desc, maptostackmap_putpage, 0, NULL },
	{ NULL, NULL, 0, NULL },
};

