
/*
 * i405_subr.c
 *
 *	@(#)i405_subr.c	1.10 (UCLA Ficus Project) 9/26/94
 *
 * Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/param.h>
#include <sys/time.h>
#include <sys/vnode.h>
#include <sys/vfs.h>
#include <sys/errno.h>
#include <i405/i405.h>
#include <sys/user.h>


#ifdef NOTDEF
#define DODEBUG(A) A
#else
#define DODEBUG(A)
#endif


/*
 *--------------------------------------------------------------------------
 *
 * i405_getfsindex -- map from a fs string name to an index in vfstab
 *
 * DESCRIPTION:
 *	i405_getfsindex determines the equivlant of MOUNT_UFS
 *	at run-time.  Of course, why Sun just didn't
 *	define the numeric constants in the first place
 *	(instead of this string nonsense), I'll never know.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
int
i405_getfsindex(fsname)
	char *fsname;
{
#ifdef OUTOFKERNEL
	/*
	 * Ook, these numbers don't matter because there's only
	 * one file system type in this address space.
	 */
	return 0;
#else
	int i;

	for (i=0; &vfssw[i]<vfsNVFS; i++)
		if (vfssw[i].vsw_name && strcmp(fsname, vfssw[i].vsw_name) == 0)
			return i;
	/*
	 * Since this is called only from vfs_mount,
	 * this should never happen.
	 */
	panic ("i405_getfsindex: missing file system");
#endif
}


/*
 *--------------------------------------------------------------------------
 *
 * i405_vn_einval -- just return EINVAL
 *
 *
 * DESCRIPTION:
 *	i405_vn_einval always returns einval.  You can use it to give
 *	a negative error message back about a particular vnodeop.
 *	(This is distinct from EOPNOTSUPP which implies a configuration
 *	error.)
 *
 * SEE ALSO:
 *	i405_vn_default
 *
 * RETURN VALUES:
 *      Always returns EINVAL.
 *
 *--------------------------------------------------------------------------
 */
int
i405_vn_einval()
{
	return EINVAL;
}


/*
 *--------------------------------------------------------------------------
 *
 * i405_vn_default -- just return EOPNOTSUPP
 *
 *
 * DESCRIPTION:
 *	i405_vn_default generic "default" routine that just returns an error.
 *	Use it to block operations you don't want to bypass.
 *
 * RETURN VALUES:
 *      Always returns EOPNOTSUPP.
 *
 *--------------------------------------------------------------------------
 */
int
i405_vn_default()
{
	return EOPNOTSUPP;
}


/*
 *--------------------------------------------------------------------------
 *
 * i405_vn_noerror -- just return 0
 *
 *
 * DESCRIPTION:
 *	i405_vn_noerror is a generic vnode routine which simply returns zero.
 *	Use it to block operations you don't want to bypass.
 *
 * RETURN VALUES:
 *      Always returns 0.
 *
 *--------------------------------------------------------------------------
 */
int
i405_vn_noerror()
{
	return 0;
}



/*
 *--------------------------------------------------------------------------
 *
 * i405_vn_cmp -- default vnode compare routine
 *
 *
 * DESCRIPTION:
 *	i405_vn_cmp is a simple vnode compare routine.
 *	If the pointers match, so do the vnodes.
 *
 * RETURN VALUES:
 *      True/false if the vnodes match.
 *
 *--------------------------------------------------------------------------
 */
int
i405_vn_cmp(ap)
	struct nvop_cmp_args *ap;
{
	return (ap->a_vp1 == ap->a_vp2);
}



/*
 *--------------------------------------------------------------------------
 *
 * i405_generate_unique_fsid -- given a vfsp and a per-vfs index, make a fsid
 *
 *
 * DESCRIPTION:
 *	Given a vfsp and a per-vfs index, make a fsid.
 *	This routine highly OS release dependent.
 *
 *
 * RETURN VALUES:
 *      Returns an fsid.
 *
 *--------------------------------------------------------------------------
 */
long
i405_generate_unique_fsid(vfsp, mntno)
	struct vfs *vfsp;
	int mntno;
{
	return 0x0ffff & (long)makedev(vfs_fixedmajor(vfsp), (0xff & mntno));
}



/*
 *--------------------------------------------------------------------------
 *
 * old_vn_rele -- release a vnode, calling the *old* VOP_INACTIVE
 *
 *
 * DESCRIPTION:
 *	This function implements VN_RELE on a vnode.
 *	If the reference count drops to zero it calls the *old* style
 *	VOP_INACTIVE, so this can be used with non-converted file systems.
 *
 * SEE ALSO:
 *	new_vn_rele
 *
 * RETURN VALUES:
 *      None.
 *
 *--------------------------------------------------------------------------
 */
void
old_vn_rele(vp)
	register struct vnode *vp;
{
        if ((vp)->v_count == 0)
                panic("old_vn_rele");
	/* This silly call method is to go through the old routine. */
        if (--(vp)->v_count == 0)
		(*(vp)->v_op->vn_inactive)((vp),u.u_cred);
}

/*
 *--------------------------------------------------------------------------
 *
 * new_vn_rele -- release a vnode, perhaps calling NVOP_INACTIVE
 *
 *
 * DESCRIPTION:
 *	This function implements VN_RELE on a vnode.
 *	If the reference count drops to zero it calls the *new* style
 *	VOP_INACTIVE, so this can be used on only file systems
 *	supporting the new interface.
 *
 * SEE ALSO:
 *	old_vn_rele
 *
 * RETURN VALUES:
 *      None.
 *
 *--------------------------------------------------------------------------
 */
void
new_vn_rele(vp)
	register struct vnode *vp;
{
	USES_NVOP_INACTIVE;
        if ((vp)->v_count == 0)
                panic("new_vn_rele");
        if (--(vp)->v_count == 0)
                NVOP_INACTIVE((vp), u.u_cred);
}




/*
 *--------------------------------------------------------------------------
 *
 * struuid -- convert an nca_uuid to a string (for printing)
 *
 *
 * DESCRIPTION:
 *	Convert an nca_uuid to a string (for printing).
 *	The string buffer should be at least 40 bytes long.
 *	No check is made for string overflow.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
void
struuid(s, uuid)
	char *s;
	struct nca_uuid *uuid;
{
#ifdef NOTDEF
	sprintf(s, "uuid(%8x%4x.%2x.%2x.%2x.%2x.%2x.%2x.%2x.%2x)",
		uuid->time_high, uuid->time_low >> 16,
		uuid->id[0],   /* is family */
		uuid->id[1],
		uuid->id[2],
		uuid->id[3],
		uuid->id[4],
		uuid->id[5],
		uuid->id[6],
		uuid->id[7]);
#else
	/*
	 * Sunos 4.1.1 kernel sprintf doesn't understand %<digit>x
	 */
	sprintf(s, "uuid(%x,%x.%x.%x.%x.%x.%x.%x.%x.%x)",
		(unsigned)uuid->time_high, (unsigned)(uuid->time_low >> 16),
		uuid->id[0],   /* is family */
		uuid->id[1],
		uuid->id[2],
		uuid->id[3],
		uuid->id[4],
		uuid->id[5],
		uuid->id[6],
		uuid->id[7]);
#endif
}




#define UUID_HASH_BUCKETS 0x20
#define UUID_HASH_MASK (UUID_HASH_BUCKETS-1)
#define UUID_TO_HASH(UUID) ((UUID)->time_high & UUID_HASH_MASK)

#define UUIDEQ(A,B) (bcmp((A),(B),sizeof(struct nca_uuid))==0)

struct vnodeop_desc *uuid_hash[UUID_HASH_BUCKETS];

/*
 *--------------------------------------------------------------------------
 *
 * i405_uuid_to_desc -- map from a uuid to a vnodeop_desc structure
 *

 * DESCRIPTION:
 *	Map from a uuid to a vnodeop_desc structure.
 *	This function is used for cross-address space calls
 *	(from out-of-the kernel or through NFS) to identify operations.
 *
o *	This function now does hashing, so repeated calls should
 *	be reasonably fast.
 *
 * RETURN VALUES:
 *      Returns a the vnodeop_desc or NULL if the uuid can't be found.
 *
 *--------------------------------------------------------------------------
 */
struct vnodeop_desc *
i405_uuid_to_desc (uuidp)
	struct nca_uuid *uuidp;		/* IN: operation uuid */
{
	int i;
	struct vnodeop_desc *vdp = NULL, *nvdp;
	int hash = UUID_TO_HASH(uuidp);

#ifdef INTGEN_TWO
#else /* INTGEN_TWO */
	if (OLDUUID(uuidp))
		panic("i405_uuid_to_desc: received old smalluuid.");
#endif /* INTGEN_TWO */

	DODEBUG(printf ("looking up uuid %d, hash=%d\n", uuid->time_high, hash));
	DODEBUG(printf ("vdp = %x\n", uuid_hash[hash]));

	/*
	 * First check the hash chain.
	 * We assert that UUID-less ops can't get put on the hash chain.
	 */
	vdp = uuid_hash[hash];
	while (vdp) {
		DODEBUG(printf ("vdp=%x\n", vdp));
		DODEBUG(printf ("ucla_transport=%x\n", UCLA_TRANSPORT(vdp)));
		DODEBUG(printf ("ucla_transport->vdut_uuid_p=*%x\n", UCLA_TRANSPORT(vdp)->vdut_uuid_p));
#ifdef INTGEN_TWO
		if (UUIDEQ(vdp->vdesc_uuid_p,uuidp)) {
			return vdp;
		}
		else vdp = vdp->vdesc_uuidhash;
#else
		if (UUIDEQ(UCLA_TRANSPORT(vdp)->vdut_uuid_p,uuidp)) {
			return vdp;
		}
		else vdp = UCLA_TRANSPORT(vdp)->vdut_uuidhash;
#endif
	};
	DODEBUG(printf ("item not found."));

	/*
	 * Hash miss.  See if it's there with a linear search.
	 */
	for (i=0; i405_op_descs[i]; i++) {
		nvdp = i405_op_descs[i];
		DODEBUG(printf ("nvdp=%x\n", nvdp));
#ifdef INTGEN_TWO
		DODEBUG(printf ("vdp->vdesc_uuid_p=*%x\n", nvdp->vdesc_uuid_p));
		/* Skip UUID-less ops. */
		if (nvdp->vdesc_uuid_p == NULL)
			continue;
		/* Check it. */
		if (UUIDEQ(nvdp->vdesc_uuid_p, uuidp)) {
			/*
			 * Found it.  Add to end of hash chain.
			 */
			nvdp->vdesc_uuidhash = NULL;
			if (uuid_hash[hash]==NULL) {
				uuid_hash[hash] = nvdp;
				return nvdp;
			} else {
				vdp = uuid_hash[hash];
				while (nvdp->vdesc_uuidhash)
					vdp = nvdp->vdesc_uuidhash;
				return nvdp->vdesc_uuidhash = nvdp;
			};
		};
#else
		DODEBUG(printf ("ucla_transport=%x\n", UCLA_TRANSPORT(nvdp)));
		DODEBUG(printf ("ucla_transport->vdut_uuid_p=*%x\n", UCLA_TRANSPORT(nvdp)->vdut_uuid_p));
		if (UUIDEQ(UCLA_TRANSPORT(nvdp)->vdut_uuid_p,uuidp)) {
			/*
			 * Found it.  Add to end of hash chain.
			 */
			UCLA_TRANSPORT(nvdp)->vdut_uuidhash = NULL;
			if (uuid_hash[hash]==NULL) {
				uuid_hash[hash] = nvdp;
				return nvdp;
			} else {
				vdp = uuid_hash[hash];
				while (UCLA_TRANSPORT(vdp)->vdut_uuidhash)
					vdp = UCLA_TRANSPORT(vdp)->vdut_uuidhash;
				return UCLA_TRANSPORT(vdp)->vdut_uuidhash = nvdp;
			};
		};
#endif
	};
	return NULL;
}



/*-
 *--------------------------------------------------------------------------
 *
 * safe_vnodeop_failure -- die a horrible death on bad vnodeops
 *
 * DESCRIPTION:
 *	Called when a vnodeop is invoked on a null vnode
 *	(if SAFE_VNODEOPS is defined by the caller).
 *
 * RETURN VALUES:
 *      Never returns (but declared as int to be used in function calls).
 *
 *--------------------------------------------------------------------------
 */
int
safe_vnodeop_failure()
{
	panic("safe_vnodeop_failure: vnode operation invoked on null vnode.");
}


static int (**i405_dead_vnodeop_p)();
static struct vnodeopv_entry_desc i405_dead_vnodeop_entries[] = {
	{ &nvop_default_desc, i405_vn_einval },
	{ NULL, NULL }
};
struct vnodeopv_desc i405_dead_vnodeop_opv_desc =
	{ &i405_dead_vnodeop_p, i405_dead_vnodeop_entries };

struct vnode *i405_dead_vp;

/*-
 *--------------------------------------------------------------------------
 *
 * i405_init_dead_vp -- build a vnode operation sink vnode
 *
 * DESCRIPTION:
 * 	i405_dead_vnode is a vnode which returns errors for all operations.
 *	Use it as a sink for bad things
 * 	and as a "null" vnode where a null vnode pointer would cause
 * 	bad operation invocations.
 *
 * RETURN VALUES:
 *      void.
 *
 *--------------------------------------------------------------------------
 */
void
i405_init_dead_vp()
{
	static struct vnode dead_vnode;

	i405_dead_vp = &dead_vnode;
	VN_INIT(i405_dead_vp, NULL, VNON, 0);
	i405_dead_vp->v_op = (struct vnodeops *)(i405_dead_vnodeop_p);
}
