
/*
 * i405_svcm.c
 *
 *	@(#)i405_svcm.c	1.4 (UCLA Ficus Project) 10/25/93
 *
 * Copyright (c) 1991-1993 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * i405_svcm.c
 *
 * Stackable vnode cache manager code for the new interface.
 *
 * This code is original, except for the hash functions
 * at the end which come from SunOS 4.1 ufs.  -JSH 7-Oct-91
 *
 * The cache manager provides a mechanism to reigster page
 * ranges with vnodes to keep cache consistency in a multi-
 * layer stack.
 *
 *
 * External interface:
 *
 * int svcm_namevp (vp): return the "name" structure
 *     for a given vnode.  (allocate a name, if required)
 * int svcm_unregister (ownvp, cred): destroy any page ownership for
 *     ownvp (syncs and throws away all pages).
 * int svcm_register (ownvp, off, len, cred): ownvp acquires ownership
 *     for all pages in the range [off,off+len).  May sleep to
 *     sync pages to backing store if they're stolen from other layers.
 *     Creds are used to write pages to disk.
 *
 * int svcm_name_lock(name): lock the page-cache of the stack containing vp
 * int svcm_name_unlock(name): unlock the stack
 * (Macros svcm_{,un}lock provide the interface with vnodes.)
 *
 *
 * Locking policy:
 *
 * The vnode stack must be locked (with svcm_name_lock) before
 * beginning code which might modify the page cache.
 * Once the stack lock has been acquired, any per-layer
 * locking can be done.
 *
 * Vop_{get,put}page should acquire this lock.
 * Vop_new{get,put}page assume that the stack lock has already 
 * been acquired.
 *
 */



#ifdef JOHNH

#include <sys/param.h>
#include <sys/time.h>
#include <sys/vnode.h>
#include <i405/i405.h>
#include <newofs/newo_interface.h>
#include <sys/buf.h>
#include <sys/kmem_alloc.h>
#include <sys/proc.h>
#include <sys/systm.h>   /* for NOMEMWAIT() */
#if 0
#include <sys/buf.h>
#include <sys/uio.h>
#include <sys/vmmeter.h>
#include <sys/vmsystm.h>
#include <sys/mman.h>
#include <sys/vfs.h>
#include <sys/debug.h>
#include <sys/trace.h>
#include <sys/ucred.h>

#include <vm/hat.h>
#include <vm/as.h>
#include <vm/seg.h>
#include <vm/rm.h>
#include <vm/pvn.h>
#include <vm/page.h>
#include <vm/seg_map.h>
#endif



#if 1
#define DODEBUG(A) if (svcm_printfs) A;
#else
#define DODEBUG(A)
#endif

#define ASSERT(EX) if (!(EX)) { printf("ASSERTION failed: %s, file: %s, line: %d\n", "EX", __FILE__, __LINE__); }
#define ASSERTF(EX,FLAG) if (!(EX)) { (FLAG)=1; printf("ASSERTION failed: %s, file: %s, line: %d\n", "EX", __FILE__, __LINE__); }

#define SANITY_CHECKS
#define REMEMBER_OWNER



int svcm_disable = 0;
int svcm_printfs = 0;

#define MAX_SEG_LEN ((u_int)0xffffffff)

/*
 * Avoid subroutine call every time we need to steal a page.
 */
#define SVCM_ACQUIRE_SEG(OWNVP, SEG, OFF, LEN, CRED) \
	(((SEG)->ownvp == (OWNVP)) ? 0 : \
	svcm_acquire_seg(OWNVP, SEG, OFF, LEN, CRED))


/*
 * Segments are the unit to keep track of things.
 * Segments are doulbly linked lists, NULL terminated at the ends.
#ifndef CIRCLE
 * NEEDSWORK: Circularly linked lists would be simplier.
#endif
 * There is always at least one segment present.
 */
struct svcm_seg {
	struct svcm_seg *prev, *next;
	u_int beg, end;
	struct vnode *ownvp;
};

struct svcm_name {
	struct svcm_name *prev, *next;  /* in the name hash */
	struct vnode *namevp;           /* the name */
	struct svcm_seg *seg;           /* head of segs */
 	u_int wanted:1;			/* someone wants to lock */
	int lockcount;			/* count of people holding the lock */
	long locker;			/* who owns the lock */
};





/***********************************************************************
 *
 * LOCKING CODE
 *
 */

/*
 * Implement a simple locking scheme.
 *
 * NEEDSWORK: Right now this is quite simple.
 * For example, it prohibits concurrency when we're doing a putpage
 * to flush a page.  This is quite bad, since a putpage implies I/O
 * implies slow.  Of course, this is no worse the SunOS already
 * had it, so there we are.
 *
 * SVCM locking supports nesting so that both vop_rdwr and vop_getpage
 * can concurrently lock the object.
 */
#define SVCM_NAME_LOCK(N) svcm_name_lock(N)
#define SVCM_NAME_UNLOCK(N) svcm_name_unlock(N)
#define SVCM_ISLOCKED(N) ((N)->locked)


int
svcm_name_lock (name)
	struct svcm_name *name;
{
	long wanter = uniqpid();

	DODEBUG(printf("svcm_name_lock: %x\n", name));
	if (name->lockcount && NOMEMWAIT())  /* refuse to sleep the pageout daemon */
		return EWOULDBLOCK;
	if (name->lockcount && name->locker != wanter) {
		name->wanted = 1;
		DODEBUG(printf("svcm_name_lock: %x sleeping\n", name));
		sleep ((caddr_t)name, PINOD);
	};
	name->lockcount++;
	name->locker = wanter;
	return 0;
}


int
svcm_name_unlock (name)
	struct svcm_name *name;
{
	DODEBUG(printf("svcm_name_unlock: %x\n", name));
	if (--name->lockcount < 0)
		panic ("svcm_name_unlock: unlocked node.");
#ifdef REMEMBER_OWNER
	name->locker = NULL;
#endif
	if (name->lockcount==0 && name->wanted) {
		name->wanted = 0;
		DODEBUG(printf("svcm_name_unlock: %x waking others\n", name));
		wakeup((caddr_t)name);
	};
	return 0;
}





/***********************************************************************
 *
 * DEBUGGING CODE
 *
 */


void
svcm_dump_pages(name)   /* debug routine */
	struct svcm_name *name;
{
	struct svcm_seg *segi = name->seg;
	int c;

	c = '[';
	do {
		printf ("%c%x-%x:%x", c, segi->beg, segi->end, segi->ownvp);
		c = ',';
		segi = segi->next;
	} while (segi);
	printf ("]\n");
}


#ifdef SANITY_CHECKS
static void
breakhere()
{
}

static void
svcm_verify_name(name)   /* debug routine */
	struct svcm_name *name;
{
	struct svcm_seg *segi = name->seg;
	int fail = 0;

	/*
	 * seglist must be
	 *   (1) complete (all bytes covered)
	 *   (2) properly linked
	 *   (3) uncompressable (every adjacent seg has different ownvp)
	 */

	ASSERTF(segi->prev == NULL, fail);			 /* (2) */
	ASSERTF(segi->beg == 0, fail);				 /* (1) */
	while (segi->next) {
		ASSERTF(segi->end+1 == segi->next->beg, fail);	 /* (1) */
		ASSERTF(segi == segi->next->prev, fail);	 /* (2) */
		ASSERTF(segi->ownvp != segi->next->ownvp, fail); /* (3) */
		segi = segi->next;
	};
	/* redundant with loop: ASSERTF(segi->next == NULL, fail);  (2) */
	ASSERTF(segi->end == MAX_SEG_LEN, fail);		 /* (1) */
	if (fail)
		breakhere();
}
#endif




/***********************************************************************
 *
 * HASHING CODE
 *
 */

static struct svcm_name *svcm_name_list = NULL;

/*
 * Find a name on our hash tables.
 * NEEDSWORK: Should really be a hash table, not just a linked list.
 */
static struct svcm_name *
svcm_name_find (vp)
	struct vnode *vp;
{
	struct svcm_name *name;
	for (name=svcm_name_list; name; name=name->next)
		if (name->namevp == vp)
			return name;
	return NULL;
}


/*
 * put a name on our hash queues.
 */
static void
svcm_enhash_name(name)
	struct svcm_name *name;
{
	name->next = svcm_name_list;
	name->prev = NULL;
	if (name->next)
		name->next->prev = name;
	svcm_name_list = name;
}

static void
svcm_unhash_name (name)
	struct svcm_name *name;
{
	if (name->prev) {
		name->prev->next = name->next;
	} else {
		svcm_name_list = name->next;
	};
}



/***********************************************************************
 *
 * MEMORY ALLOCATION CODE
 *
 */

/*
 *
 * Boring memory stuff.
 *
 */

/* could sleep */
static struct svcm_name *svcm_name_base = NULL;

static struct svcm_name *
svcm_name_alloc()
{
	struct svcm_name *name =
		(struct svcm_name*)new_kmem_fast_alloc(&svcm_name_base,
				   sizeof(struct svcm_name), 0x10, KMEM_SLEEP);
	if (name)
		bzero(name, sizeof(struct svcm_name));
	return name;
}

static void
svcm_name_free (name)
	struct svcm_name *name;
{
	kmem_fast_free (&svcm_name_base, name);
}

/* could sleep */
static struct svcm_seg *svcm_seg_base = NULL;

static struct svcm_seg *
svcm_seg_alloc()
{
	struct svcm_seg *seg =
		(struct svcm_seg*)new_kmem_fast_alloc(&svcm_seg_base,
				   sizeof(struct svcm_seg), 0x10, KMEM_SLEEP);
	if (seg)
		bzero(seg, sizeof(struct svcm_seg));
	return seg;
}

static void
svcm_seg_free (seg)
	struct svcm_seg *seg;
{
	kmem_fast_free (&svcm_seg_base, seg);
}



/***********************************************************************
 *
 * NAMING CODE
 *
 */


/*
 * Create an svcm context for a given vp.
 *
 * We correctly handle the possibility of multiple invocations,
 * just like when you allocate new inodes.  (We could sleep on memory
 * allocation.  When we come back, someone else could have created
 * us.)
 *
 */
int
svcm_namevp (vp, cred)
	struct vnode *vp;
	struct ucred *cred;
{
	struct vnode *namevp;
	int error;
	USES_NVOP_CACHENAMEVP;
	struct svcm_name *name;

	if (svcm_disable)
		return 0;

	if (error = NVOP_CACHENAMEVP(vp, &namevp, cred))
		return error;
	DODEBUG(printf ("svcm_namevp: cachenamevp for %x is %x\n", vp, namevp));
	
	while (NULL == (name = svcm_name_find (namevp))) {
		struct svcm_name *name2 = svcm_name_alloc();
		struct svcm_seg *seg = svcm_seg_alloc();

		if (name2 == NULL || seg==NULL) {
			panic ("svcm_namevp: out of memory.\n");
		};

		if (NULL != (name = svcm_name_find(namevp))) {
			/*
			 * Someone created it before us.
			 * Free ours, take theirs.
			 */
			svcm_name_free(name2);
			svcm_seg_free(seg);
		} else {
			/*
			 * Ours is the new name.
			 * Make the default segment.
			 * Add it to the hash lists.
			 *
			 * Note optimization:
			 * the creator gets whole file ownership by default.
			 */
			name = name2;
			name->prev = name->next = NULL;
			name->namevp = namevp;
			name->seg = seg;
			name->lockcount = name->wanted =
				name->locker = 0;
			seg->next = seg->prev = NULL;
			seg->beg = 0;
			seg->end = MAX_SEG_LEN;
			seg->ownvp = vp;
			svcm_enhash_name(name);
			DODEBUG(printf("svcm_namevp: creating new name %x\n", name));
		};
	};
	if (vp->v_svcm_name && vp->v_svcm_name != name)
		panic ("svcm_namevp: vp %x already named.\n", vp);
	vp->v_svcm_name = name;
	DODEBUG(printf("svcm_namevp: naming %x (real %x) with %x\n", vp, namevp, name));
#ifdef SANITY_CHECKS
	svcm_verify_name(name);
#endif
	return 0;
}




/***********************************************************************
 *
 * SEGMENT MANIPULATION CODE
 *
 */

/*
 * svcm_acquire_seg: force pages to disks.
 */
static int
svcm_acquire_seg(ownvp, seg, off, len, cred)
	struct vnode *ownvp;
	struct svcm_seg *seg;
	u_int off, len;
	struct ucred *cred;
{
	u_int putoff, putlen;
	int error = 0;
	USES_NVOP_PUTPAGE;

	if (seg->ownvp == ownvp)   /* already own it */
		return 0;

	DODEBUG(printf("svcm_acquire_seg(%x,%x,%x,%x,%x): putpage called\n", 
		       ownvp, seg, off, len, cred));
	/*
	 * To acquire it, we just call putpage
	 * to sync the page(s) to backing storage.
	 *
	 * The correct range to put is the minimum of this seg
	 * and [off, off+len).
	 */
	putoff = MAX(off, seg->beg);
	if (seg->beg == 0 && seg->end == MAX_SEG_LEN)   /* avoid overflow problem */
		putlen = len;
	else putlen = MIN(len, seg->end - seg->beg + 1);
	error = NVOP_PUTPAGE(seg->ownvp, putoff, putlen, B_INVAL, cred);
	return error;
}

/*
 * svcm_seg_split: split a segment into two.
 * Each of the new segs are optionally returned.
 */
static void
svcm_seg_split(seg, place, seg1p, seg2p)
	struct svcm_seg *seg;
	u_int place;
	struct svcm_seg **seg1p, **seg2p;
{
	struct svcm_seg *seg2 = svcm_seg_alloc();
	if (!seg2)
		panic("svcm_seg_split: null seg return.");
	*seg2 = *seg;   /* structure assign */
	seg->next = seg2;
	seg2->prev = seg;
	seg->end = place-1;
	seg2->beg = place;
	if (seg2->next)
		seg2->next->prev = seg2;
	if (seg1p)
		*seg1p = seg;
	if (seg2p)
		*seg2p = seg2;
}

/*
 * svcm_seg_merge -- merge a seg with its following seg (maybe).
 */
static int
svcm_seg_merge (seg)   /* returns boolean "if merge occurred" */
	struct svcm_seg *seg;
{
#if 1
	if (!seg)
		panic ("svcm_seg_merge: NULL segment.\n");
#endif
	if (seg->next == NULL)
		return 0;
	DODEBUG(printf ("svcm_seg_merge: seg=%x, next=%x, nextnext=%x\n",
			seg, seg->next, seg->next->next));
	if (seg->ownvp == seg->next->ownvp) {
		struct svcm_seg *nextseg = seg->next;
		/*
		 * Do a merge.
		 */
		seg->end = nextseg->end;
		seg->next = nextseg->next;
		if (nextseg->next)
			nextseg->next->prev = seg;
		svcm_seg_free(nextseg);
		return 1;
	};
	return 0;
}


/*
 * Transfer ownership to new vp from current owner of [off,off+len).
 *
 * Assumes single access to name & its seg list.
 */
static int
svcm_grabseg (ownvp, name, off, len, cred)
	struct vnode *ownvp;
	struct svcm_name *name;
	u_int off, len;
	struct ucred *cred;
{
	struct svcm_seg *segi = name->seg;
	struct svcm_seg *seg_beg, *seg_end;
	u_int beg = off;
	u_int end = off+len-1;
	int error;

	DODEBUG(printf ("svcm_grabseg (%x,%x, off %x,len %x,%x): beg=%x, end=%x\n",
			ownvp, name, off, len, cred, beg, end));
	/*
	 * Find place to start allocing.
	 */
	for (;;) {
		if (beg >= segi->beg && beg <= segi->end)
			break;
		segi = segi->next;
		if (segi==name->seg)
			panic ("svcm_grabseg: seg (beg) not found.");
	};
	seg_beg = segi;

	/*
	 * Now go through segments, making them owned by the right vp.
	 *
	 * If we get an error acquiring a segment, we just bail.
	 * This may leave invalid page ownership because of split/merge
	 * stuff which we haven't done yet.  This is OK, because
	 * presumably the new vp will not try to allocate pages
	 * because it will abort with this error.
	 */
	for (;;) {
		if (end <= segi->end)   /* we're done */
			break;
		if (error = SVCM_ACQUIRE_SEG(ownvp, segi, off, len, cred))
			return error;
		segi = segi->next;
		if (segi==name->seg)
			panic ("svcm_grabseg: seg (end) not found.");
	};
	if (end >= segi->beg) {   /* get the last bit of the last seg */
		if (error = SVCM_ACQUIRE_SEG(ownvp, segi, off, len, cred))
			return error;
	};
	seg_end = segi;

	/*
	 * Now split or merge seg's, as required.
	 * (Only if ownership must changes.)
	 */
	if (beg != seg_beg->beg && seg_beg->ownvp != ownvp) {
		int seg_end_matches = (seg_beg == seg_end);
		svcm_seg_split(seg_beg, beg, NULL, &seg_beg);
		if (seg_end_matches)
			seg_end = seg_beg;
	};
	if (end != seg_end->end && seg_end->ownvp != ownvp) {
		svcm_seg_split(seg_end, end+1, &seg_end, NULL);
	};
	/*
	 * Make one single new segment with
	 * everything owned by ownvp,
	 * then go through and free up any previously unused segments.
	 *
	 * We do this my making seg_beg the new segment
	 * and throwing away everything after that.
	 */
	seg_beg->end = seg_end->end;
	seg_beg->ownvp = ownvp;   /* give the new segment to its owner */
	/* Seg_beg is now complete and good.  Free remaining segs. */
	if (seg_beg != seg_end) {
		segi = seg_beg->next;
		while (segi != seg_end) {
			struct svcm_seg *nextseg;
			nextseg = segi->next;
			svcm_seg_free (segi);
			segi = nextseg;
		};
	};
	/* seg_end = NULL;   --- seg_end is no longer valid */
	/*
	 * Finally, try to merge new segement with its neighbors.
	 */
	svcm_seg_merge(seg_beg);   /* merge new seg with following seg */
	if (seg_beg->prev) {
		svcm_seg_merge(seg_beg->prev);   /* and preceeding seg */
		/* seg_beg = NULL;  --- seg_beg is no longer valid */
	};

	return 0;
}





/***********************************************************************
 *
 * REGISTER/UNREGISTER CODE
 *
 */

/*
 * We assume that we can always alloc names,
 * and that the won't ever go away out from
 * under us because they go away only when the
 * naming vnode goes away and that won't happen
 * because our caller hold a reference to it.
 *
 * Calling vp (ownvp) must be locked.
 *
 * Calls to svcm_register should be pushed as low in
 * the getpage call chain as possible to avoid this
 * overhead if at all possible.  For the null layer,
 * this means that svcm_register is called in page fill portion of
 * the newgetapage routine of the lowest layer---we can place it there
 * because we know that if we have a cached page, we must 
 * already hold ownership to that page.
 *
 * The trade-off is that by placing this
 * code in the "fill individual page" code, we can only
 * transfer ownership of pages one page at a time.
 */
int
svcm_register (ownvp, off, len, cred)
	struct vnode *ownvp;
	u_int off, len;
	struct ucred *cred;
{
	struct svcm_name *name = ownvp->v_svcm_name;
	int error;

	if (svcm_disable)
		return 0;

	if (name == NULL)
		panic ("svcm_register: null name on vp %x\n", ownvp);

	if (off==0 && len==0) {
		off = 0;
		len = MAX_SEG_LEN;
	};

#ifdef OLD_LOCK_STRATEGY
	SVCM_NAME_LOCK(name);
#endif
	DODEBUG(printf("svcm_register(%x,%x,%x,%x)\n", ownvp, off,len,cred)); 
	DODEBUG(svcm_dump_pages(name));
	if (error = svcm_grabseg (ownvp, name, off, len, cred))
		return error;
	DODEBUG(svcm_dump_pages(name));
#ifdef SANITY_CHECKS
	svcm_verify_name(name);
#endif
#ifdef OLD_LOCK_STRATEGY
	SVCM_NAME_UNLOCK(name);
#endif
	return 0;
}



/*
 * svcm_unregister:  called by vnode when it's going away.
 * Frees pages and any claims on pages.
 * Also frees name structure if we were the lowest one in the stack.
 *
 * We assume that the calling vp is locked.
 *
 * We assume that unregister cannot be called on the bottom
 * of stack vnode if there are other nodes in the stack.
 * This is always valid because other nodes must hold a refcount
 * to the bottom of stack node and so it cannot go inactive
 * and so cannot call unregister.
 */
int
svcm_unregister (vp, cred)
	struct vnode *vp;
	struct ucred *cred;
{
	struct svcm_name *name = vp->v_svcm_name;
	USES_NVOP_PUTPAGE;
	int stack_bottom;
	struct svcm_seg *segi, *segnext;
	int error = 0;

	if (svcm_disable)
		return 0;

	DODEBUG(printf("svcm_unregister(%x,%x)\n", vp, cred));

	/*
	 * No name implies that no getpages have been done,
	 * so we quit.
	 */
	if (!name)
		return 0;

	/*
	 * Sync all of our pages with the backing store.
	 */
	error = NVOP_PUTPAGE(vp, 0, 0, B_INVAL, cred);
	/*
	 * Give away all segments to the namevp
	 * (just to keep them owned by someone).
	 */
	segi = name->seg;
	stack_bottom = (vp==name->namevp);
	DODEBUG(svcm_dump_pages(name));
	DODEBUG(printf("  : stack_bottom = %d\n", stack_bottom));
	if (!stack_bottom) {
		do {
			segnext = segi->next;
			if (segi->ownvp == vp) {
				/*
				 * If we own this segment
				 * (and we're not the only one left),
				 * give it away.
				 */
				segi->ownvp = name->namevp;
				svcm_seg_merge (segi);   /* merge forward, if necessary */
				if (segi->prev) {   /* merge back */
					svcm_seg_merge (segi->prev);
					segi = NULL;  /* segi is not necessarily valid after merge */
				};
			};
			segi = segnext;
		} while (segi);
	} else {
		/*
		 * If we're the stack bottom, there should
		 * be only one segment.
		 */
		if (segi->prev != NULL &&
		    segi->next != NULL)
			panic ("svcm_unregstier: stack bottom, but multiple segments.\n");
	};
	DODEBUG(svcm_dump_pages(name));

	/*
	 * If we're the bottom,
	 * we must release the segments, and then the name.
	 * There should be only one.
	 */
	if (stack_bottom) {
		if (name->seg->prev || name->seg->next)
			panic("svcm_name_destroy: bottom of stack with multiple segments.");
		svcm_seg_free(name->seg);
		svcm_unhash_name(name);
		svcm_name_free(name);
		DODEBUG(printf("svcm_unregister: name freed\n"));
	};
#ifdef SANITY_CHECKS
	if (!stack_bottom)
		svcm_verify_name(name);
#endif
	return error;
}


#endif


