 
/*
 * newvint_syscalls.c
 *
 * User systemcalls to the new vnode interface.
 *
 * The uap's below aren't the ones that come directly from the
 * system call.  They're just pointers into user space
 * of the user's vnode args struct.
 *
 * Vnodes are given to the user level as file descriptiors.
 * This means that when a user program crashes, its vnodes
 * will be released.
 *
 * To support the null layer out-of-kernel (i.e., under nfs),
 * we need to be able to to get the vfs structure for
 * a vnode, and to do a vget on that vfs.  As a temporary
 * patch, these are new system calls.  The vfs is given
 * to the user as a pointer, but list of current mounted
 * vfs's is always checked to make sure the pointer is valid.
 *
 */
 
 
#include <sys/param.h>
#ifdef OUTOFKERNEL
#include "myuser.h"
#else
#include <sys/user.h>
#endif
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/vfs.h>
#include <sys/vnode.h>
#include <sys/nvnode.h>
#include <rpc/types.h>
#include <rpc/xdr.h>
#include <newvint/xxdr.h>
#include <sys/syslog.h>

#include <newofs/newo_interface.h>

#if 0
int newvint_syscalls_debug=1;
#define DODEBUG(A) if (newvint_syscalls_debug) { A };
#else
#define DODEBUG(A)
#endif


extern int num_operations;
#ifdef BACKVINT
extern int num_back_operations;
#endif
extern struct vnodeop_desc *vnodeop_descs[];






 
/*
 * Read the vnodeops vector configuration from the user.
 * This operation isn't really used any longer.  UUIDs
 * are passed between user and kernel.   -JSH 7-Feb-91
 */
user_vnodeop_operations (uap)
        struct a {
                int *numops;
                struct uuid_and_offset *buf;
                int maxops;
        } *uap;
{
        int i;
	int num_new_ops = num_operations;

#ifdef BACKVINT
	num_new_ops -= num_back_operations;
#endif
        u.u_error = 0;
        if (!u.u_error && uap->numops)
                u.u_error = copyout(&num_new_ops, uap->numops, sizeof(int));
	if (!u.u_error && uap->buf)
		if (uap->maxops < num_new_ops) {
			u.u_error = ENOMEM;
			return;
		} else for (i=0; vnodeop_descs[i] && !u.u_error; i++) {
			u.u_error = copyout(&vnodeop_descs[i]->vdesc_uuid,
					    &uap->buf[i].uo_uuid,
					    sizeof(uap->buf[i].uo_uuid));
			if (!u.u_error)
			    u.u_error = copyout(&vnodeop_descs[i]->vdesc_offset,
					    &uap->buf[i].uo_offset,
					    sizeof(uap->buf[i].uo_offset));
		};
}



/*
 * The system call interface to the new vnode interface.
 */
user_common_vnodeop(uap, is_old_op)
        struct a {
                int uvp;
                int uuid;
                struct nvop_generic_args **app;
        } *uap;
	int is_old_op;
{
        struct vnodeop_desc *vdescp;
        struct nvop_generic_args *xap=NULL;
        XDR xdrs;
	struct vnode *vp;
	caddr_t xdrbuf[XDRUSER_DEPTH];
	struct nvop_xdr_public public;
	extern struct vnodeop_desc *uuid_to_desc();
	struct ucred *tmpcr;
	int offset;
	extern int dummy_freefunc();
		
#ifdef OUTOFKERNEL
	mallocmap();
#endif
	u.u_error = 0;
#if 0
	/*
	 * At one time we took offsets directly into the kernel,
	 * rather than uuids.  Fortunately, this misstep has been
	 * fixed, leaving only this ugly mess of code which is 
	 * commented out.
	 */
	if (uap->uuid < 100) {
		offset = uap->uuid;   /* backwards compatibility */
		if (offset < 0 || offset > num_operations) {
			u.u_error = EINVAL;
			return;
		};
		vdescp = vnodeop_descs[offset
#ifdef BACKVINT
			       - 28  /* NEEDSWORK: gross, but commented out.
				      * This whole offset OOK thing 
				      * should go away.
				      */
#endif
				       ];
	}
	else
#endif
	{
		if (NULL==(vdescp = uuid_to_desc(uap->uuid))) {
			u.u_error = EINVAL;
			return;
		};
		offset = vdescp->vdesc_offset;
	};
	if (vdescp == &nvop_inactive_desc) {
		u.u_error = user_inactive(uap->uvp, *(uap->app));
		return;
	};
	if (vdescp->vdesc_xdr==NULL) {
		u.u_error = EOPNOTSUPP;
		return;
	};

	DODEBUG(printf ("nvop %d, vdesc=%x, name=%s\n", offset, 
			vdescp, vdescp->vdesc_name));
	DODEBUG(printf ("calling xdr_usercreate\n"));
	xdruser_create(&xdrs, uap->app, xdrbuf, XDRUSER_DEPTH, XDR_DECODE);
	public.freefunc = dummy_freefunc;
	public.mode = xfer_in;
	public.otw_to_vp = xdr_fdtovp;
	public.vp_to_otw = xdr_vptofd;
	public.vp_free = xdr_fdvpfree;
	public.otw_to_vpp = 
		public.vpp_to_otw = 
		public.vpp_free = xdr_fdvpp_stuff;
	xdrs.x_public = (caddr_t)&public;
	DODEBUG(printf("Calling xdr_alt_reference on the way in.\n"));
	DODEBUG(printf("xdrs=%x, xap=%x, size=%d, xdr_proc=%x\n", 
		       &xdrs, &xap, vdescp->vdesc_size, vdescp->vdesc_xdr));
	if (!xdr_alt_reference(&xdrs, &xap,
			       vdescp->vdesc_size, vdescp->vdesc_xdr)) {
		log (LOG_ERR, "user_common_vnodeop: xdr_alt_reference xdr error on copyin.\n");
		if (!u.u_error)  {  /* make sure errno is set */
			u.u_error = EINVAL;
			return;
		};
	};
	if (xdrs.x_private != xdrs.x_base) {
		panic ("xdr_user mismatch.");
	};
	DODEBUG(printf("Back from xdr_alt_ref\n"));
	vp = (vdescp->vdesc_flags&VDESC_INVPP) ? 
		*xap->a_un.a_vpp : xap->a_un.a_vp;
	if (!vp) {
		u.u_error = EINVAL;
		return;
	};
	DODEBUG(printf ("calling, vp=%x, err=%d \n", vp, u.u_error));

	/*
	 * What a hack.  Some file systems (like UFS)
	 * take the uid out of u.u_cred, not the cred field
	 * passed to the function.
	 *    So to make this work, we make them both the same.
	 */
	tmpcr = NULL;
	if (vdescp->vdesc_cred_offset != VDESC_NO_OFFSET) {
		tmpcr = u.u_cred;
		u.u_cred = *(struct ucred**)
			(((char*)xap)+vdescp->vdesc_cred_offset);
	};

#ifndef OUTOFKERNEL
	/*
	 * Call the routine.
	 */
	if (is_old_op)
		u.u_error = (*xap->a_desc->vdesc_newtoold)(xap);
	else u.u_error = VCALL(vp, offset, xap);
#else
	mallocmap();
	printf ("Routine called.\n");
#endif

	/*
	 * Undo the cred grossness.
	 */
	if (tmpcr) {
		u.u_cred = tmpcr;
	};

	DODEBUG(printf ("back, err=%d\n", u.u_error));
	if (!XDR_SETPOS(&xdrs, uap->app))
		u.u_error = EINVAL;
	/* reset the pointer into userland */
	xdrs.x_op = XDR_ENCODE;
	public.mode = xfer_out;
	DODEBUG(printf ("calling xdr_alt_ref for the results.\n"));
	if (!u.u_error)
		if (!xdr_alt_reference(&xdrs, &xap,
				       vdescp->vdesc_size, vdescp->vdesc_xdr))
			if (!u.u_error) {
				log (LOG_ERR, "user_common_vnodeop: xdr_alt_reference xdr error on copyin.\n");
				u.u_error = EINVAL;
			};
	DODEBUG(printf ("calling xdr_alt_ref to free the args.\n"));
	/*
	 * Free our memory.
	 */
	XDR_SETPOS(&xdrs, uap->app);/* reset the pointer into userland */
	xdrs.x_op = XDR_FREE;
	xdr_alt_reference(&xdrs, &xap, vdescp->vdesc_size,
			  vdescp->vdesc_xdr);
	XDR_DESTROY(&xdrs);
	DODEBUG(printf ("exiting, err=%d\n", u.u_error));

#ifdef OUTOFKERNEL
	mallocmap();
#endif
}
 
 
/*
 * The system call interface to the _old_ vnode interface.
 */
user_old_vnodeop(uap)
        struct a {
                int uvp;
                int offset;
                struct nvop_generic_args **ap;
        } *uap;
{
	user_common_vnodeop(uap, 1);
}

 
/*
 * The system call interface to the _new_ vnode interface.
 */
user_new_vnodeop(uap)
        struct a {
                int uvp;
                int offset;
                struct nvop_generic_args **ap;
        } *uap;
{
	user_common_vnodeop(uap, 0);
}
 


/*
 * Given a user vnode, get its vfsp and *vfsp.
 */
vnode_to_vfs (uap)
	struct a {
		int uvp;
		struct vnode **vfsp;   /* return the kernel vfsp pointer */
		struct vnode *vfs;    /* return the kernel vfsp contents */
	} *uap;
{
	struct vnode *vp, *fdtovp();

	u.u_error = 0;
	if (NULL==(vp = fdtovp(uap->uvp)))
		u.u_error = EINVAL;
	if (!u.u_error && uap->vfsp)
		u.u_error = copyout(&vp->v_vfsp,uap->vfsp,sizeof(struct vfs*));
	if (!u.u_error && uap->vfs)
		u.u_error = copyout(vp->v_vfsp,uap->vfs,sizeof(struct vfs));
}


/*
 * This is a hack since VOP_FID doesn't work the way the protocol
 * works (it wants to return a variable length of data).
 *
 * NEEDSWORK: Fix VOP_FID to return a fixed length.
 */
vnode_to_fid (uap)
	struct a {
		int uvp;
		struct fid *fid;
	} *uap;
{
	struct vnode *vp, *fdtovp();
	struct fid *fid;
	USES_NVOP_FID;
	
	u.u_error = 0;
	if (NULL==(vp = fdtovp(uap->uvp)))
		u.u_error = EINVAL;
	if (!u.u_error)
		u.u_error = NVOP_FID(vp, &fid);
#if 0   /* for debugging */
	{       int i;
		printf ("vnode_to_fid: len=%d, ", fid->fid_len);
		for (i=0; i<10; i++)
			printf ("%2x", fid->fid_data[i]);
		printf ("\n");
	}
#endif
	if (!u.u_error)
		u.u_error = copyout((caddr_t)fid,
				    (caddr_t)uap->fid,
				    fid->fid_len+sizeof(short));
}

/*
 * Given a vfsp and a fid, return the given vnode.
 */
fid_to_vnode (uap)
	struct a {
		struct vfs *vfsp;
		struct fid *fidp;
		int *uvp;
	} *uap;
{
	struct vfs *vfsp;
	struct vnode *vp;
	struct fid fid;
	int fd;

	u.u_error = 0;
	for (vfsp=rootvfs; vfsp; vfsp=vfsp->vfs_next)
		if (vfsp==uap->vfsp)
			break;
	if (!vfsp)
		u.u_error = ESTALE;
	/*
	 * This two part copy of the fid madness is necessary because
	 * it's conceivable that the user has only allocated part
	 * of the fid, so copying sizeof(struct fid) could cause
	 * a memory fault.   -JSH 8-Feb-91
	 */
	if (!u.u_error)
		u.u_error = copyin(uap->fidp, &fid, sizeof(short));
	if (!u.u_error && fid.fid_len)
		if (fid.fid_len>=0 && fid.fid_len<MAXFIDSZ)
			u.u_error = copyin(uap->fidp, &fid, sizeof(short)+fid.fid_len);
		else u.u_error = EINVAL;
	if (!u.u_error)
		u.u_error = VFS_VGET(vfsp, &vp, &fid);
	if (!u.u_error)
		fd = vptofd(vp);
	if (!u.u_error)
		u.u_error = copyout(&fd, uap->uvp, sizeof(int));
}


/*
 * Given a vfsp and a newfid, return the given vnode.
 */
newfid_to_vnode (uap)
	struct a {
		struct vfs *vfsp;
		struct newfid *newfidp;
		int *uvp;
	} *uap;
{
	struct vfs *vfsp;
	struct vnode *vp;
	struct newfid newfid;
	char *newfiddata = NULL;
	int fd;

	u.u_error = 0;
	for (vfsp=rootvfs; vfsp; vfsp=vfsp->vfs_next)
		if (vfsp==uap->vfsp)
			break;
	if (!vfsp)
		u.u_error = ESTALE;
	if (!u.u_error)
		u.u_error = copyin(uap->newfidp,&newfid,sizeof(struct newfid));
	/*
	 * Now get the fid data.
	 */
	newfid.newfid_data_space = newfid.newfid_data_elems;
	if (newfid.newfid_data_space > 1024) {
		u.u_error = ENOMEM;
		printf ("newfid_to_vnode: data too big.\n");
	};
	if (!u.u_error)
		newfiddata = kmem_alloc(newfid.newfid_data_elems);
	if (!u.u_error && newfid.newfid_data_elems) {
		u.u_error = copyin (newfid.newfid_data, newfiddata, newfid.newfid_data_elems);
		newfid.newfid_data = newfiddata;
	}
	else u.u_error = EINVAL;
	if (!u.u_error) {
		printf ("Calling newvget on vfsp=%x, newfid_elems=%d\n",
			vfsp, newfid.newfid_data_elems);
		u.u_error = VFS_NEWVGET(vfsp, &vp, NULL, &newfid);
	};
	if (!u.u_error)
		fd = vptofd(vp);
	if (!u.u_error)
		u.u_error = copyout(&fd, uap->uvp, sizeof(int));
	if (newfiddata)
		kmem_free (newfiddata, newfid.newfid_data_elems);
}

