
/*
 * utok_vnodeops.c
 *
 * A package to emulate vnodes out of kernel.
 *
 * Intercepts returned vnodes from (lookup, mkdir, create),
 * and wraps a user-level structure around in-kernel vnodes
 * (so VN_HOLD and VN_RELE work like they should).
 *
 * This has one known bug:  all utoknodes are listed as type VNON.
 * Of course, according to Rosenthal, you shouldn't check that
 * field anyway.
 */


#include <sys/param.h>
#include <sys/systm.h>
#include <sys/user.h>
#include <sys/vfs.h>
#include <sys/vfs_stat.h>
#include <sys/vnode.h>
#include <sys/nvnode.h>

#include <newvint/newvint_syscalls.h>

#ifdef KERNEL
#define mem_alloc(S) kmem_alloc(S)
#define mem_free(P,S) kmem_free(P,S)
#else
#define mem_alloc(S) malloc(S)
#define mem_free(P,S) free(P)
#endif


#if 0
#define DODEBUG(A) A
#else
#define DODEBUG(A)
#endif



/* forward defs */
struct vfs *utok_makevfs();



#define UTOKNTOVP(UTOKN) (&(UTOKN)->vnode)
#define VPTOUTOKN(VP) ((struct utok_node *)((VP)->v_data))
#define UTOKNTOLOWER(UTOKN) ((struct vnode*)((UTOKN)->lowervp))


struct utok_node {
	struct vnode *lowervp;   /* lower level vnode */
	struct vnode vnode;   /* our vnode */
	struct utok_node *next;
};



struct utok_node *utok_base_node = NULL;


int utok_bypass(),
	utok_open(),
	utok_inactive(),
	utok_fid();

int (**utok_vnodeop_p)();
struct vnodeop_defn utok_vnodeop_decls[] = {
	{ &utok_vnodeop_p, &nvop_default_desc, utok_bypass },
	{ &utok_vnodeop_p, &nvop_open_desc, utok_open },
	{ &utok_vnodeop_p, &nvop_inactive_desc, utok_inactive },
	{ &utok_vnodeop_p, &nvop_fid_desc, utok_fid },
	{ NULL, NULL, NULL }
};



/*****************************************************************************
 * What follows should be in utok_subr.c
 *****************************************************************************/

struct utok_node *
utok_find (lowervp)
	struct vnode *lowervp;
{
	struct utok_node *up = utok_base_node;

	while (up)
		if (up->lowervp == lowervp) {
			VN_HOLD(UTOKNTOVP(up));
			return up;
		}
		else up = up->next;
	return NULL;
}

void
utok_free (old_up)
	struct utok_node *old_up;
{
	struct utok_node *up = utok_base_node, *pup=NULL;

	if (up==old_up) {
		utok_base_node = up->next;
	} else while (up)
		if (up == old_up) {
			pup->next = up->next;
			break;
		}
		else {
			pup = up;
			up = up->next;
			if (up==NULL)
				panic ("utok_free: up not found.");
		};
	mem_free (up, sizeof(struct utok_node));
}


int
makeutoknode (lowervp, vpp)
	struct vnode *lowervp, **vpp;
{
	struct utok_node *utokn;
	static struct vfs *utok_vfsp = NULL;

	if (utokn = utok_find(lowervp)) {
		DODEBUG(printf ("utoknode found.\n"));
		*vpp = UTOKNTOVP(utokn);
		return 0;
	};

	utokn = (struct utok_node*)mem_alloc(sizeof(struct utok_node));
	
	if (utokn==NULL)
		return ENOMEM;

	/* make a vfs structure, if necessary. */
	if (utok_vfsp==NULL) {
		utok_vfsp = utok_makevfs(lowervp);
	};

/* VN_INIT(VP, VFSP, TYPE, DEV)	*/
	VN_INIT(UTOKNTOVP(utokn), utok_vfsp, VNON, 0);

	utokn->lowervp = lowervp;
	utokn->next = utok_base_node;
	utok_base_node = utokn;

	UTOKNTOVP(utokn)->v_op = (struct vnodeops *) utok_vnodeop_p;
	UTOKNTOVP(utokn)->v_data = (caddr_t)utokn;
	*vpp = UTOKNTOVP(utokn);
	return 0;
}



/*****************************************************************************
 * What follows should be in utok_vfsops.c
 *****************************************************************************/


int
utok_error()
{
	panic ("Bad utok vfs operation called.\n");
}

/*
 * Convert to an in-kernel vget.
 * vfs_data is the in-kernel vfs structure.
 */
int
utok_vget(vfsp, vpp, fidp, newfidp)
	struct vfs *vfsp;
	struct vnode **vpp;
	struct fid *fidp;
	struct newfid *newfidp;
{
	extern int errno;
	int error;

	if (fidp) {
		if (-1==syscall(SYS_fid_to_vnode,vfsp->vfs_data, fidp, vpp))
			return errno;
	} else {
		if (-1==syscall(SYS_newfid_to_vnode, vfsp->vfs_data, newfidp, vpp))
			return errno;
	};

	return makeutoknode (*vpp, vpp);   /* wrap a utoknode around things. */
}

static struct vfsops utok_vfsops = {
	utok_error,   /* mount */
	utok_error,   /* unmount */
	utok_error,   /* root */
	utok_error,   /* statfs */
	utok_error,   /* sync */
	utok_vget,    /* vget */
	utok_error,   /* mountroot */
	utok_error,   /* swapvp */
};

struct vfs *
utok_makevfs(lowervp)
	struct vnode *lowervp;
{
	struct vfs *vfsp, *lowervfsp;

        if (-1==syscall(SYS_vnode_to_vfs, lowervp, &lowervfsp, NULL))
		return NULL;

	vfsp = (struct vfs *)mem_alloc(sizeof(struct vfs));
	if (vfsp==NULL)
		return NULL;

	bzero(vfsp, sizeof(struct vfs));
	vfsp->vfs_op = &utok_vfsops;
	vfsp->vfs_bsize = 8*1024;
	vfsp->vfs_data = (caddr_t)lowervfsp;

	return vfsp;
}





/*****************************************************************************
 * What follows should be in utok_vnodeops.c
 *****************************************************************************/



/*
 * utok_open does some grossness to take care of the _only_
 * case where the first parameter is a vpp.
 */
int
utok_open(ap)
	struct nvop_open_args *ap;
{
	int error;
	struct vnode **old_vpp;
	struct vnode *old_lower_vp, *new_lower_vp;
	
	old_vpp = ap->a_vpp;   /* save old value of vpp */

	/*
	 * Open could change the vpp, so we give the lower-level
	 * a new place to put the pointer.
	 */
	new_lower_vp = old_lower_vp = UTOKNTOLOWER(VPTOUTOKN(*ap->a_vpp));
	ap->a_vpp = &new_lower_vp;

	error = VKCALL(*ap->a_vpp, VOFFSET(nvop_open), ap);

	if (!error) {
		/*
		 * If the lower-pointer changed, build a new utoknode
		 * around it.  Otherwise just replace the old
		 * contents of vpp.
		 */
		if (new_lower_vp != old_lower_vp)
			return makeutoknode(new_lower_vp, ap->a_vpp);
	};
	/*
	 * If error or no change in vpp, put it back.
	 */
	ap->a_vpp = old_vpp;
	return error;
}


/*
 * utok_fid must be done separately because the fid defintion is 
 * bad.  It expects one side to have a pre-allcoated buffer,
 * which isn't kosher with the into the kernel XDR code.
 *
 * Until all instances of utok_fid have been changed, we
 * hack around it like this.
 */
int
utok_fid(ap)
	struct nvop_fid_args *ap;
{
	int error;
	struct fid *fidp;
	extern int errno;
	int lowervp = (int)UTOKNTOLOWER(VPTOUTOKN(ap->a_vp));
	
	if (NULL==(fidp = (struct fid*) kmem_alloc(sizeof(struct fid))))
		return ENOMEM;
	if (-1==syscall (SYS_vnode_to_fid, lowervp, fidp)) {
		kmem_free(fidp, sizeof(struct fid));
		return errno;
	};
	/*
	 * fidp->fid_len = MAXFIDSZ;  /* so freefid works *\
	 * This can't be done, because then when the fid is passed
	 * back into the kernel, it will be different.
	 * Instead, we hack freefid (in the ook version)
	 * to always free a whole fid structure.
	 *
	 * What should happen is that the maximum fid size
	 * should be figured dynamically (like vnodeop vectors)
	 * and placed in a kernel global variable.
	 * - JSH, 8-Feb-91
	 */
	*(ap->a_fidpp) = fidp;
	return 0;
}


int
utok_inactive(ap)
	struct nvop_inactive_args *ap;
{
	int error;
	struct vnode *old_vp;
	
	old_vp = ap->a_vp;
	ap->a_vp = UTOKNTOLOWER(VPTOUTOKN(ap->a_vp));
	error = VKCALL(ap->a_vp, VOFFSET(nvop_inactive), ap);

	utok_free(VPTOUTOKN(old_vp));

	return error;	
}


int
utok_bypass(ap)
	struct nvop_generic_args *ap;
{
	int error;
	struct vnode *old_vp1, *old_vp2;
	struct vnode **vp1p, **vp2p, ***vppp;

	if ((ap->a_desc->vdesc_flags & VDESC_INVPP) != 0)
		panic ("invpp not handled yet.\n");

	/*
	 * Map the vnodes going in.
	 */
	if (ap->a_desc->vdesc_vp1_offset != VDESC_NO_OFFSET) {
		vp1p=XtOffsetTo(struct vnode**,ap->a_desc->vdesc_vp1_offset,ap);
		old_vp1 = *vp1p;
		*vp1p = UTOKNTOLOWER(VPTOUTOKN(*vp1p));
	};
	if (ap->a_desc->vdesc_vp2_offset != VDESC_NO_OFFSET) {
		vp2p=XtOffsetTo(struct vnode**,ap->a_desc->vdesc_vp2_offset,ap);
		old_vp2 = *vp2p;
		*vp2p = UTOKNTOLOWER(VPTOUTOKN(*vp2p));
	};

	/*
	 * Either vp1 or vp2 have handled converting ap->a_un.a_vp 
	 * to a lower level vnode (love that pointer aliasing).
	 */
	error = VKCALL(ap->a_un.a_vp, ap->a_desc->vdesc_offset, ap);

	/*
	 * Maintain the illusion of call-by-value.
	 */
	if (ap->a_desc->vdesc_vp1_offset != VDESC_NO_OFFSET)
		*vp1p = old_vp1;
	if (ap->a_desc->vdesc_vp2_offset != VDESC_NO_OFFSET)
		*vp2p = old_vp2;
	/*
	 * Map the possible out-going vpp.
	 */
	if (ap->a_desc->vdesc_vpp_offset != VDESC_NO_OFFSET && !error) {
		vppp=XtOffsetTo(struct vnode***,
				 ap->a_desc->vdesc_vpp_offset,ap);
		error = makeutoknode(**vppp, *vppp);
	};

	return error;
}




void
utok_purge_cache(vp)
	struct vnode *vp;
{
        syscall(SYS_knfs_purge_caches, VPTOUTOKN(vp)->lowervp);
}

