
/*-
 * svcm_byterange.c
 *
 *	@(#)svcm_byterange.c	1.14 (UCLA Ficus Project) 7/17/95
 *
 * Copyright (c) 1993 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#ifdef SVCM_TWO

#include <sys/param.h>
#include <sys/time.h>
#include <sys/systm.h>   /* for NOMEMWAIT() */
#include <sys/vnode.h>
#include <i405/i405.h>
#include <i405/SVCM/svcm.h>
#include <sys/errno.h>
#include <sys/buf.h>   /* for B_INVAL */
#include <sys/debug.h>  /* for ASSERT */
#include <vm/page.h>

#if 0
#define DODEBUG(A) A
#else
#define DODEBUG(A)
#endif



/**********************************************************************
 *
 * SANITY CHECKS.
 *
 * See svcm_byterange.h for details.
 *
 */
#ifdef JOHNH_SVCM_BYTERANGE_VERIFY
#define JOHNH_SVCM_BYTERANGE_VERIFY_ASSERT(X) ASSERT(X)
#define SVCM_VERIFY_SEGLIST(SEG) svcm_verify_seglist(SEG);
#else /* ! JOHNH_SVCM_BYTERANGE_VERIFY */
#define JOHNH_SVCM_BYTERANGE_VERIFY_ASSERT(X) 
#define SVCM_VERIFY_SEGLIST(SEG) 
#endif /* JOHNH_SVCM_BYTERANGE_VERIFY */

#ifdef JOHNH_DATA_PAGE_LIST_VERIFY
#define SVCM_MARK_STABLE(NAME) ((NAME)->lock->unstable = 0)
#define SVCM_MARK_UNSTABLE(NAME) ((NAME)->lock->unstable = 1)
#define SVCM_VERIFY_SEGLIST_DATA_PAGES(NAME) /*svcm_verify_seglist_data_pages(NAME)*/
#else /* ! JOHNH_DATA_PAGE_LIST_VERIFY */
#define SVCM_MARK_STABLE(NAME)
#define SVCM_MARK_UNSTABLE(NAME)
#define SVCM_VERIFY_SEGLIST_DATA_PAGES(NAME)
#endif /* JOHNH_DATA_PAGE_LIST_VERIFY */

#ifdef JOHNH_PAGEFLIP_VERIFY
#define SVCM_VERIFY_VP_PAGE_OWNERSHIP(VP,NAME) svcm_verify_vp_page_ownership(VP,NAME)
#else /* ! JOHNH_PAGEFLIP_VERIFY */
#define SVCM_VERIFY_VP_PAGE_OWNERSHIP(VP,NAME) 
#endif /* JOHNH_PAGEFLIP_VERIFY */



/**********************************************************************
 *
 * BYTERANGE STATISTICS
 *
 */

struct svcm_byterange_stats {
	/* Allocation statistics. */
	int sbs_seg_alloc;
	int sbs_seg_free;
	/* Fast-path information. */
	int sbs_register_fastpath_hit;
	int sbs_register_fastpath_miss;
	int sbs_acquire_already_owned;
	int sbs_acquire_was_unregistered;
	int sbs_acquire_callback_required;
};

struct svcm_byterange_stats svcm_byterange_stats[SVCM_OBJ_CLASSES];

#ifdef SVCM_BYTERANGE_KEEP_STATS
#define SVCM_BYTERANGE_STATS_INC(CLASS,FIELD) ((svcm_byterange_stats[CLASS].FIELD)++)
#define SVCM_BYTERANGE_STATS_ADD(CLASS,FIELD,COUNT) (svcm_byterange_stats[CLASS].FIELD += (COUNT))
#else /* ! SVCM_KEEP_STATS */
#define SVCM_BYTERANGE_STATS_INC(CLASS,FIELD)
#define SVCM_BYTERANGE_STATS_ADD(CLASS,FIELD,COUNT)
#endif /* SVCM_KEEP_STATS */





/**********************************************************************
 *
 * SEGMENT DEFINITIONS
 *
 */

static struct svcm_byterange_name default_range = { 0, MAX_SEG_LEN };




/***********************************************************************
 *
 * MEMORY ALLOCATION CODE
 *
 */

/* could sleep */
static struct svcm_seg *svcm_seg_base = NULL;
#define SVCM_SEG_INC 0x10

static struct svcm_seg *
svcm_seg_alloc()
{
	struct svcm_seg *seg =
		(struct svcm_seg*)new_kmem_fast_alloc(&svcm_seg_base,
				   sizeof(struct svcm_seg), SVCM_SEG_INC, KMEM_SLEEP);
	SVCM_BYTERANGE_STATS_ADD(0, sbs_seg_alloc, SVCM_SEG_INC);
	if (seg)
		bzero(seg, sizeof(struct svcm_seg));
	return seg;
}

static void
svcm_seg_free (seg)
	struct svcm_seg *seg;
{
	SVCM_BYTERANGE_STATS_INC(0, sbs_seg_free);
	kmem_fast_free ((caddr_t *) &svcm_seg_base, (caddr_t)seg);
}



/***********************************************************************
 *
 * SEGMENT MANIPULATION CODE
 *
 */


/*-
 *--------------------------------------------------------------------------
 *
 * svcm_acquire_seg -- acquire rights to a byterange
 *
 * DESCRIPTION:
 *	Acquire rights to a byterange, calling back any parties
 *	who currently hold it cached if necessary.  When we call
 *	them back they will write their pages to "backing store",
 *	or with page flipping they'll flip them down and then we
 *	flip them back up.
 *
 *	Note: may leave the seg list unnormalized.
 *
 *	With SVCM_PAGEFLIP, calls to this routine may change
 *	the page list of (theoretically) any layer.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
svcm_acquire_seg(name, own_vp, seg, obj_class, range, status, callback_made, cred)
	struct svcm_name *name;		/* IN: invlolved name */
	struct vnode *own_vp;		/* IN: vp that wants the pages */
	struct svcm_seg *seg;		/* IN: segment containing range */
	int obj_class;			/* IN */
	struct svcm_byterange_name *range;   /* IN: range to acquire */
	enum svcm_status status;	/* IN: desired rights */
	int *callback_made;		/* OUT: set iff a callback was made,
					 * unchanged otherwise. */
	struct ucred *cred;		/* IN */
{
	USES_NVOP_CACHE_CALLBACK;
	USES_NVOP_PUTPAGE;
	int error;

	/* Check to see if we already own it the right way. */
	if (SEG_MATCHES(seg, own_vp) && seg->status == status) {
		SVCM_BYTERANGE_STATS_INC(obj_class, sbs_acquire_already_owned);
		return 0;
	}

	/*
	 * If it's uncached, just take it.
	 * We take the whole thing as an optimization.
	 *
	 * Note: This may unnormalize the segment list.
	 *
	 * Note: We don't need to callback the owner if it was uncached.
	 * In fact, the owner might no longer exist.
	 */
	if (seg->status == SVCM_STATUS_UNCACHED) {
		seg->own_vp = own_vp;
		seg->status = status;
		SVCM_BYTERANGE_STATS_INC(obj_class, sbs_acquire_was_unregistered);
		return 0;
	};

	/*
	 * To acquire it we call back the vp.
	 *
	 * The correct range to put is the minimum of this seg
	 * and [off, off+len).
	 */
#ifdef SVCM_PAGEFLIP
	/*
	 * The callback below syncronizes pages with backing store.
	 * If we're doing page flipping, then the pages will be
	 * left in the bottom stack layer and page ownership/location
	 * information temporarily disagrees.
	 */
	if (status == SVCM_STATUS_EXCLUSIVE)
		*callback_made = 1;
#endif /* SVCM_PAGEFLIP */
	SVCM_BYTERANGE_STATS_INC(obj_class, sbs_acquire_callback_required);
	error = NVOP_CACHE_CALLBACK(seg->own_vp, name, obj_class,
				   sizeof(struct svcm_byterange_name),
				   (void*)range, cred);

#ifdef SVCM_PAGEFLIP
	/*
	 * HACK ALERT!
	 * With page flipping we need extra code after 
	 * the callback.  We should make a seperate
	 * function, but I don't.  When this code is used
	 * by classes we'll need to do that work.
	 */
	if (obj_class != SVCM_CLASS_DATA_PAGES)
		panic("svcm_acquire_seg: pageflip hack NEEDSWORK.");
	/*
	 * Ownership change has flipped the pages to the stack bottom.
	 * We're not done yet, though:  this configuration violates
	 * our constraints that pages only be placed in layers
	 * that own them.
	 *
	 * Fix things up now by flipping them to their proper owner.
	 *
	 *
	 * SECONDARY HACK ALERT:
	 * We assume below that pages flip to name->vp.
	 * This assumption isn't guaranteed with complex stacks
	 * and it treats names as non-opaque.
	 *
	 *
	 * Bug: There was a problem in this code (found 950707).
	 * We're getting pages in the bottom of the stack which are
	 * not flipped back up.  To leave pages stranded in
	 * half-flipped state violates our constraints and potentially
	 * ends up causing duplicated pages.
	 *
	 * The bug occurs because the cache callback (above) rounds
	 * sync'ed pages to fit the blocksize when it flips pages down
	 * the stack.  When we flip pages back up the stack here,
	 * we only flip the orignal request, not the rounded request.
	 *
	 * We fix this bug by having pvn_stackkluster do rounding
	 * so we're now guaranteed that the callback doesn't do any
	 * additional rounding.
	 */
	if (status == SVCM_STATUS_EXCLUSIVE) {
		/* Don't try to flip pages to ourselves. */
		if (name->vp != own_vp)
			(void) svcm_pageflip_byterange(name,
				    name->vp, own_vp,
				    range->offset,
				    range->length);
	};
	/*
	 * UNCACHED at bottom of stack means that we must write data
	 * to stable storage.
	 */
	if (status == SVCM_STATUS_UNCACHED) {
		/*
		 * We now allow only the bottom layer to
		 * do uncached owns.
		 */
		ASSERT(name->vp == own_vp);
		/*
		 * Fast path: skip work if no data remaining.
		 */
		if (own_vp->v_pages != NULL) {
			(void) NVOP_PUTPAGE(own_vp, 0, 0, B_INVAL, (struct ucred *)0);
			ASSERT(own_vp->v_pages == NULL);
		};
	};
#endif /* SVCM_PAGEFLIP */

	return error;
}

/*
 * svcm_seg_split: split a segment into two.
 * Each of the new segs are optionally returned.
 */
static void
svcm_seg_split(seg, place, seg1p, seg2p)
	struct svcm_seg *seg;
	off_t place;
	struct svcm_seg **seg1p, **seg2p;
{
	struct svcm_seg *seg2 = svcm_seg_alloc();

	if (!seg2)
		panic("svcm_seg_split: null seg return.");
	*seg2 = *seg;   /* structure assign */
	seg->next = seg2;
	seg2->prev = seg;
	seg->end = place-1;
	seg2->beg = place;
	if (seg2->next)
		seg2->next->prev = seg2;
	if (seg1p)
		*seg1p = seg;
	if (seg2p)
		*seg2p = seg2;
}

/*
 * svcm_seg_merge -- merge a seg with its following seg (maybe).
 */
static int
svcm_seg_merge (seg)   /* returns boolean "if merge occurred" */
	struct svcm_seg *seg;
{
	JOHNH_SVCM_BYTERANGE_VERIFY_ASSERT(seg != NULL);
	if (seg->next == NULL)
		return 0;
	DODEBUG(printf ("svcm_seg_merge: seg=%x, next=%x, nextnext=%x\n",
			seg, seg->next, seg->next->next));
	if (SEG_MATCHES(seg, seg->next->own_vp)) {
		struct svcm_seg *nextseg = seg->next;
		/*
		 * Do a merge.
		 */
		seg->end = nextseg->end;
		seg->next = nextseg->next;
		if (nextseg->next)
			nextseg->next->prev = seg;
		svcm_seg_free(nextseg);
		return 1;
	};
	return 0;
}



/*-
 *--------------------------------------------------------------------------
 *
 * svcm_data_pages_init -- 
 *
 * DESCRIPTION:
 *	Init data pages.  By default, make them all uncached.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
int
svcm_data_pages_init(name, own_vp, obj_class, cred)
	struct svcm_name *name;
	struct vnode *own_vp;
	int obj_class;
	struct ucred *cred;
{
	struct svcm_seg **base_seg_p =
		(struct svcm_seg **)&name->classes[obj_class];
	struct svcm_seg *seg;

	if (*base_seg_p)
		return 0;

	/*
	 * Allocate a new segment.
	 */
	seg = svcm_seg_alloc();
	seg->next = seg->prev = NULL;
	seg->beg = 0;
	seg->end = MAX_SEG_END;
	/*
	 * Award all pages to the current layer.
	 */
	seg->own_vp = own_vp;
	seg->status = SVCM_STATUS_UNCACHED;

	*base_seg_p = seg;

	return 0;
}

/*-
 *--------------------------------------------------------------------------
 *
 * svcm_byterange_register -- acquire/release rights to a byterange
 *
 * DESCRIPTION:
 *	Svcm_byterange_register is the central routine in stack cache mangement.
 *	A group of vnodes participating in a cache (identified by {\em name})
 *	cooperate.  When a given vnode (identified by {\em own_vp})
 *	desires a particular byterange {\em range}
 *	of type {\em obj_class}, it calls this routine.
 *	When it returns (own_vp) is guaranteed {\em status}
 *	rights to the object.
 *
 *	To accomplish this, svcm_byterange_register keeps track of who has
 *	what rights to each object.  If currently held rights are
 *	incompatible with an incoming requests, callbacks are made
 *	on existing objects asking them to release their holdings.
 *
 *	svcm_byterange_register assumes that the cache lock object is
 *	already held by the calling process.
 *
 *	Allowed status are SVCM_STATUS_UNCACHED and SVCM_STATUS_EXCLUSIVE.
 *	SVCM_STATUS_SHARED and SVCM_STATUS_NONCACHABLE are not supported.
 *
 * SEE ALSO:
 *	svcm_unregister.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
int
svcm_byterange_register(name, own_vp, obj_class, range_len, range, status, cred)
	struct svcm_name *name;		/* IN: stack name */
	struct vnode *own_vp;		/* IN: desired owning vp */
	enum svcm_obj_classes obj_class;/* IN: cache subobject name */
	int range_len;			/* IN: length of obj_name */
	struct svcm_byterange_name *range; /* IN: cache subobject item */
	enum svcm_status status;	/* IN: desired status */
	struct ucred *cred;		/* IN */
{
	struct svcm_seg *base_seg, *segi, *nextseg;
	struct svcm_seg *seg_beg, *seg_end;
	off_t beg;
	off_t end;
	int error = 0;
	int callback_made;

	/*
	 * Fast-path the common case (at least for SVCM_CLASS_DATA_PAGES):
	 * we want exclusive access to a new page and
	 * we have exclusive access to the whole file.
	 *
	 * Sigh.  Have to check base_seg->{prev,next} in case
	 * there are watch points.
	 */
	base_seg = (struct svcm_seg*)(name->classes[obj_class]);
	if ((status == SVCM_STATUS_EXCLUSIVE) &&
	    (base_seg &&
	     base_seg->prev == NULL && base_seg->next == NULL &&
	     base_seg->beg == 0 && base_seg->end == MAX_SEG_END &&
	     base_seg->own_vp == own_vp &&
	     base_seg->status == SVCM_STATUS_EXCLUSIVE)) {
		SVCM_BYTERANGE_STATS_INC(obj_class, sbs_register_fastpath_hit);
		return 0;
	} else {
		SVCM_BYTERANGE_STATS_INC(obj_class, sbs_register_fastpath_miss);
	};
		
	/*
	 * Handle default range.
	 */
	if (range == NULL) {
		range_len = sizeof(default_range);
		range = &default_range;
	};

	if (range->length == 0)   /* zero length segment -> no work to do */
		return 0;
	beg = range->offset;
	end = range->offset + (range->length - 1);

	JOHNH_SVCM_BYTERANGE_VERIFY_ASSERT(name != NULL);
	JOHNH_SVCM_BYTERANGE_VERIFY_ASSERT(!(obj_class == SVCM_CLASS_DATA_PAGES && range->length % PAGESIZE != 0));

	/*
	 * Don't manage directories.
	 */
	if (own_vp->v_type == VDIR)
		return 0;

	/*
	 * Byterange doesn't currently support all statuses.
	 */
	if (status != SVCM_STATUS_UNCACHED && status != SVCM_STATUS_EXCLUSIVE)
		panic("svcm_byterange_register: called with invalid status.");


#ifdef SVCM_PAGEFLIP
	/*
	 * When pageflipping only the bottom layer can register
	 * as uncached.  If other layers want to register uncached
	 * we turn the call into exclusive access by the bottom layer.
	 * We must do this because the pages will end up there.
	 */
	if (status == SVCM_STATUS_UNCACHED && own_vp != name->vp) {
		own_vp = name->vp;
		status = SVCM_STATUS_EXCLUSIVE;
	};
#else /* ! SVCM_PAGEFLIP */
	/*
	 * If we're making it uncached, then own_vp is irrelevant.
	 * Make it null here so (1) we don't use it by mistake,
	 * (2) we can merge adjacent uncached segments.
	 */
	if (status == SVCM_STATUS_UNCACHED)
		own_vp = NULL;
#endif /* SVCM_PAGEFLIP */

	SVCM_MARK_UNSTABLE(name);
	/*
	 * Possible initialization required.
	 */
	if (base_seg == NULL) {
		if (error = SVCM_INIT(name, own_vp, obj_class, cred)) {
			SVCM_MARK_STABLE(name);
			return error;
		};
		base_seg = (struct svcm_seg*)name->classes[obj_class];
	};

	/*
	 * Find place to start allocing.
	 */
	for (segi = base_seg;;) {
		if (beg >= segi->beg && beg <= segi->end)
			break;
		segi = segi->next;
		if (!segi)
			panic ("svcm_byterange_register: seg (beg) not found.");
	};
	seg_beg = segi;

	/*
	 * Now go through segments, making them owned by the right vp.
	 *
	 * If we get an error acquiring a segment, we just bail.
	 * This may leave invalid page ownership because of split/merge
	 * stuff which we haven't done yet.  This is OK, because
	 * presumably the new vp will not try to allocate pages
	 * because it will abort with this error.
	 */
	callback_made = 0;
	for (;;) {
		if (end <= segi->end)   /* we're done */
			break;
		if (error = svcm_acquire_seg(name, own_vp, segi, obj_class, range, status, &callback_made, cred))
			goto out;
		segi = segi->next;
		if (!segi)
			panic ("svcm_byterange_register: seg (end) not found.");
	};
	if (end >= segi->beg) {   /* get the last bit of the last seg */
		if (error = svcm_acquire_seg(name, own_vp, segi, obj_class, range, status, &callback_made, cred))
			goto out;
	};
	seg_end = segi;

	/*
	 * Now split or merge seg's, as required.
	 * (Only if ownership must changes.)
	 */
	if (beg != seg_beg->beg && !SEG_MATCHES(seg_beg, own_vp)) {
		int seg_end_matches = (seg_beg == seg_end);
		svcm_seg_split(seg_beg, beg, NULL, &seg_beg);
		if (seg_end_matches)
			seg_end = seg_beg;
	};
	if (end != seg_end->end && !SEG_MATCHES(seg_end, own_vp)) {
		svcm_seg_split(seg_end, end+1, &seg_end, NULL);
	};
	/*
	 * Make one single new segment with
	 * everything owned by own_vp,
	 * then go through and free up any previously unused segments.
	 *
	 * We do this my making seg_beg the new segment
	 * and throwing away everything after that.
	 */
	seg_beg->end = seg_end->end;
	seg_beg->own_vp = own_vp;   /* give the new segment to its owner */
	seg_beg->status = status;
	/* Seg_beg is now complete and good.  Free remaining segs. */
	if (seg_beg != seg_end) {
		/* Walk the list and free each seg. */
		segi = seg_beg->next;
		for (;;) {
			nextseg = segi->next;
			svcm_seg_free(segi);
			if (segi == seg_end)
				break;
			segi = nextseg;
		};
		/* Now repair list pointers. */
		seg_beg->next = nextseg;
		if (nextseg)
			nextseg->prev = seg_beg;
	};
	/* seg_end = NULL;   --- seg_end is no longer valid */
	/*
	 * Finally, try to merge new segement with its neighbors.
	 */
	(void) svcm_seg_merge(seg_beg);   /* merge new seg with following seg */
	if (seg_beg->prev) {
		(void) svcm_seg_merge(seg_beg->prev);   /* and preceeding seg */
	};
	/* seg_beg = NULL;   --- seg_beg is no longer valid */

#ifdef SVCM_PAGEFLIP
	/*
	 * If a callback was made then we could have moved pages around.
	 * Detect and report that here (with EAGAIN)
	 * to allow the caller to possibly re-scan their page list.
	 */
	if (callback_made)
		error = EAGAIN;
#endif /* SVCM_PAGEFLIP */

 out:
	SVCM_MARK_STABLE(name);
	SVCM_VERIFY_SEGLIST(base_seg);
	SVCM_VERIFY_SEGLIST_DATA_PAGES(base_seg);
	/* Check page ownership of the stack base and the new owner. */
	SVCM_VERIFY_VP_PAGE_OWNERSHIP(name->vp, name);
	SVCM_VERIFY_VP_PAGE_OWNERSHIP(own_vp, name);
	return error;
}


#endif SVCM_TWO
