
/*-
 * svcm_main.c
 *
 *	@(#)svcm_main.c	1.17 (UCLA Ficus Project) 1/9/95
 *
 * Copyright (c) 1993 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


/*-
 *
 * svcm_main.c
 *
 * This is a second take on the svcm routines.
 * Here we provide the most generic of services:
 * a name database.
 *
 */


#ifdef SVCM_TWO


#include <sys/param.h>
#include <sys/time.h>
#include <sys/systm.h>   /* for NOMEMWAIT() */
#include <sys/user.h>	 /* for u.* in NOMEMWAIT() */
#include <sys/vnode.h>
#include <sys/errno.h>	 /* for EWOULDBLOCK */
#include <sys/debug.h>   /* for ASSERT */
#include <i405/i405.h>
#include <i405/SVCM/svcm.h>


#if 0
#define DODEBUG(A) A
#else
#define DODEBUG(A)
#endif



/*
 * The sanity checks here are small and can be left on all the time.
 */
#define SANITY_CHECKS
#ifdef JOHNH4C
#define SVCM_KEEP_STATS
#endif /* JOHNH4C */


struct svcm_stats {
	int lock_alloc;
	int lock_create;
	int lock_lock;
	int lock_sleep;
	int lock_destroy;
	int name_alloc;
	int name_create;
	int name_duplicate_create;
	int name_reuse;
	int name_rebirth_sleeps;
	int name_rebirth_wakes;
	int name_destroy;
};
struct svcm_stats svcm_stats;
#ifdef SVCM_KEEP_STATS
#define SVCM_STATS_INC(FIELD) ((svcm_stats.FIELD)++)
#define SVCM_STATS_ADD(FIELD,COUNT) (svcm_stats.FIELD += (COUNT))
#else /* ! SVCM_KEEP_STATS */
#define SVCM_STATS_INC(FIELD)
#define SVCM_STATS_ADD(FIELD,COUNT)
#endif /* SVCM_KEEP_STATS */



static int
svcm_unsupported_init()
{
	panic("svcm_upsupported_init: cache manager call on unsupported obj_class.");
	/*NOTREACHED*/
}


struct svcm_class_ops svcm_class_ops[SVCM_OBJ_CLASSES] = {
{ svcm_data_pages_init, svcm_byterange_register },
{ svcm_unsupported_init, svcm_byterange_register },
{ svcm_namedobj_init, svcm_namedobj_register },
{ svcm_namedobj_init, svcm_namedobj_register },
/* { svcm_unsupported_init, svcm_unsupported_init }, */
};






/***********************************************************************
 *
 * LOCKING CODE
 *
 */

static struct svcm_lock *svcm_lock_base = NULL;

static struct svcm_lock *
svcm_lock_alloc()
{
	struct svcm_lock *lock;

	SVCM_STATS_ADD(lock_alloc, (svcm_lock_base == NULL ? 0x10 : 0));
	lock = (struct svcm_lock*)new_kmem_fast_alloc(&svcm_lock_base,
				   sizeof(struct svcm_lock), 0x10, KMEM_SLEEP);
	if (lock)
		bzero(lock, sizeof(struct svcm_lock));
	return lock;
}


/*-
 *--------------------------------------------------------------------------
 *
 * svcm_lock_create -- create an unlocked lock
 *
 * DESCRIPTION:
 *	Create a lock and return it.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *	Lock is unlocked on return.
 *
 *--------------------------------------------------------------------------
 */
int
svcm_lock_create(lockp)
	struct svcm_lock **lockp;    /* OUT: lock (will be unlocked) */
{
	struct svcm_lock *lock = svcm_lock_alloc();

	if (lock == NULL)
		panic ("svcm_lock_create: out of memory.");
	lock->lockcount = lock->wanted =
		lock->locker = 0;
	lock->refcnt = 1;
	*lockp = lock;
	SVCM_STATS_INC(lock_create);
	return 0;
}

void
svcm_lock_destroy(lock)
	struct svcm_lock *lock;
{
	if (lock->refcnt < 0)
		panic("svcm_lock_destroy: rele of unref'ed lock.");
	if (lock->wanted)
		panic("svcm_lock_destroy: rele of lock with waiters on it.");
	SVCM_STATS_INC(lock_destroy);
	kmem_fast_free((caddr_t *) &svcm_lock_base, (caddr_t)lock);
}

int
svcm_lock(lock)
	struct svcm_lock *lock;
{
	long wanter = uniqpid();

	if (lock->lockcount && NOMEMWAIT())  /* refuse to sleep the pageout daemon */
		return EWOULDBLOCK;
	SVCM_STATS_INC(lock_lock);
	if (lock->lockcount && lock->locker != wanter) {
		/* printf("svcm_lock: %x sleeping\n", lock); */
		lock->wanted = 1;
		SVCM_STATS_INC(lock_sleep);
		sleep ((caddr_t)lock, PINOD);
		/* printf("svcm_lock: %x waking\n", lock); */
	};
	lock->lockcount++;
	lock->locker = wanter;
	return 0;
}

int
svcm_unlock(lock)
	struct svcm_lock *lock;
{
	if (--lock->lockcount < 0)
		panic ("svcm_unlock: unlocked node.");
	if (lock->lockcount == 0) {
#ifndef REMEMBER_OWNER
		lock->locker = NULL;
#endif
		if (lock->wanted) {
			lock->wanted = 0;
			wakeup((caddr_t)lock);
		};
	};
	return 0;
}

int
svcm_islocked(lock)
	struct svcm_lock *lock;
{
	return (lock->lockcount != 0);
}


#if 0
void
svcm_lock_check_fail(file, line)
	char *file;
	int line;
{
	extern int noprintf;
	noprintf = 0;
	panicstr = "";
	printf ("svcm_lock_check_fail: file: %s, line %d\n", file, line);
	panicstr = 0;
	panic("svcm_lock_check_fail");
}
#endif




/***********************************************************************
 *
 * HASHING CODE
 *
 */

#define NAME_HASH_BUCKETS 0x40
#define NAME_HASH_MASK (NAME_HASH_BUCKETS-1)
#define NAME_TO_HASH(NAME) ((((u_int)(NAME)) >> 4) & NAME_HASH_MASK)

static struct svcm_name *svcm_name_hash[NAME_HASH_BUCKETS];


/*-
 *--------------------------------------------------------------------------
 *
 * svcm_name_find -- locate an existing name indexed by cachevp
 *
 * DESCRIPTION:
 *	Find a cached name as indexed by cachevp.
 *	If found, bump its refcnt and return it.
 *
 *	NEEDSWORK: Should really use a hash table, not just a linked list.
 *
 *	Svcm_name_find is an internal interface.
 *	External callers should use svcm_name_create.
 *
 *	We handle a race condition on names that are going away:
 *	if they're dieing, wait for them to die and then try again.
 *
 * RETURN VALUES:
 *      Returns refcnt-bumped name or NULL if none found.
 *
 *--------------------------------------------------------------------------
 */
static struct svcm_name *
svcm_name_find(vp)
	struct vnode *vp;
{
	int hash = NAME_TO_HASH(vp);
	struct svcm_name *name;
 again:
	for (name = svcm_name_hash[hash]; name; name = name->next)
		if (name->vp == vp) {
			if (name->flags & SVCM_NAME_DIEING) {
				name->flags |= SVCM_NAME_REBORN;
				SVCM_STATS_INC(name_rebirth_sleeps);
				/* printf ("Pid %d sleeping on name %x\n", u.u_procp->p_pid, name); */
				sleep((caddr_t)name, PINOD);
				goto again;
			};
			svcm_name_hold(name);
			return name;
		};
	return NULL;
}


/*
 * put a name on our hash queues.
 */
static void
svcm_enhash_name(name)
	struct svcm_name *name;
{
	int hash = NAME_TO_HASH(name->vp);

	name->next = svcm_name_hash[hash];
	name->prev = NULL;
	if (name->next)
		name->next->prev = name;
	svcm_name_hash[hash] = name;
}

static void
svcm_unhash_name (name)
	struct svcm_name *name;
{
	if (name->prev) {
		name->prev->next = name->next;
	} else {
		int hash = NAME_TO_HASH(name->vp);
		svcm_name_hash[hash] = name->next;
	};
	if (name->next) {
		name->next->prev = name->prev;
	};
}



/***********************************************************************
 *
 * NAMING CODE
 *
 */

static struct svcm_name *svcm_name_base = NULL;

static struct svcm_name *
svcm_name_alloc()
{
	struct svcm_name *name;

	SVCM_STATS_ADD(name_alloc, (svcm_name_base == NULL ? 0x10 : 0));
	name = (struct svcm_name*)new_kmem_fast_zalloc(&svcm_name_base,
				   sizeof(struct svcm_name), 0x10, KMEM_SLEEP);
	return name;
}

static void
svcm_name_free (name)
	struct svcm_name *name;
{
	kmem_fast_free ((caddr_t *) &svcm_name_base, (caddr_t)name);
}


/*-
 *--------------------------------------------------------------------------
 *
 * svcm_name_create -- create an svcm context for a given cachevp
 *
 * DESCRIPTION:
 *
 *	Create an svcm context for a given cachevp,
 *	or locate a suitable existing name and bump its refcnt.
 *	Returns with the lock held.
 *
 *	We correctly handle the possibility of multiple invocations,
 *	just like when you allocate new inodes.  (We could sleep on memory
 *	allocation.  When we come back, someone else could have created
 *	us.)
 *
 * 	Note: On entry *namep must be NULL.  This is to catch
 *	repeated allocation bugs.  Eventually this requirement
 *	may be relaxed.
 *
 * SEE ALSO:
 *	svcm_name_find, svcm_name_destroy.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *	The name lock is locked on return.
 *
 *--------------------------------------------------------------------------
 */
int
svcm_name_create (namep, cachevp, old_lock, cred)
	struct svcm_name **namep;	/* OUT: new name (must point to a null pointer) */
	struct vnode *cachevp;		/* IN: vp describing the "stack" */
	struct svcm_lock *old_lock;	/* IN: lock to use or NULL to make one */
	struct ucred *cred;		/* IN */
{
	int error;
	struct svcm_name *name2;
	struct svcm_name *name;

	SVCM_SFO_NOT_CONVERTED;

	if (NULL == (name = svcm_name_find(cachevp))) {
		/*
		 * Fill in name, possibly sleeping in the process.
		 */
		name = svcm_name_alloc();
		name->vp = cachevp;
#if 0
		/* These fields are now zeroed when the object is freed. */
		name->prev = name->next = NULL;   /* not needed */
		bzero(name->classes, sizeof(name->classes));
		name->refcnt = 0;
		name->flags = 0;
#endif
		svcm_name_hold(name);   /* bump refcnt */
		if (old_lock) {
			/* If old lock, it must be locked already. */
			ASSERT(old_lock->lockcount > 0);
			name->lock = old_lock;
			svcm_lock_hold(name->lock);  /* one hold per name */
		} else {
			/*
			 * Create an unlocked lock.
			 * The calls svcm_enhash_name and svcm_lock
			 * must be atomic to avoid creation races.
			 */
			if (error = svcm_lock_create(&name->lock)) {
				svcm_name_free(name);
				return error;
			};
			svcm_lock(name->lock);
		};
		/*
		 * We're about to register the name.
		 * Make sure that someone hasn't beat us to the punch.
		 */
		if (name2 = svcm_name_find(cachevp)) {
			SVCM_STATS_INC(name_duplicate_create);
			/* Someone was faster.  Use their work. */
			/* Free up the almost-complete name and its lock. */
			svcm_lock_rele(name->lock);
			svcm_name_free(name);
			/* Take the new name and lock it. */
			name = name2;
			SVCM_STATS_INC(name_reuse);
			svcm_lock(name->lock);
		} else {
			/* Publish the new name atomicly after find failure. */
			SVCM_STATS_INC(name_create);
			svcm_enhash_name(name);
			DODEBUG(printf("svcm_name_create: created new name %x\n", name));
		};
	} else {
		/* If we got it off the queue, we need to lock it. */
		SVCM_STATS_INC(name_reuse);
		svcm_lock(name->lock);
	};
	
	/* Name must be locked. */
	ASSERT(name->lock->lockcount > 0);
	if (*namep)
		panic ("svcm_name_create: name already allocated.");
	*namep = name;
	return 0;
}


/*-
 *--------------------------------------------------------------------------
 *
 * svcm_name_rele -- release a reference count on a name
 *
 * DESCRIPTION:
 *	Toss a refcnt for vnode vp.  If all gone, call svcm_name_destroy.
 *	As a side-effect all callbacks promosed to vnode vp are 
 *	relinquished (and therefore all resources owned by that vp
 *	are disposed of).
 *
 *	This operation can sleep.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
int
svcm_name_rele (name, vp, cred)
	struct svcm_name *name;		/* IN: locked name */
	struct vnode *vp;		/* IN: vp no longer needing refcnt */
	struct ucred *cred;		/* IN */
{
	int i, error, saved_error = 0;

#ifdef SANITY_CHECKS
	if (!svcm_islocked(name->lock))
		panic("svcm_name_rele: name is unlocked.");
#endif /* SANITY_CHECKS */

	/*
	 * First release any cached objects.
	 * (Could sleep.)
	 * While this is happening, the object could be re-acquired
	 * and its reference count bumped.  This is OK.
	 */
	for (i = 0; i < SVCM_OBJ_CLASSES; i++)
		if (name->classes[i]) {
			error = SVCM_REGISTER(name, vp,
					      (enum svcm_obj_classes) i,
					      0, NULL,
					      SVCM_STATUS_UNCACHED, cred);
			if (error == EAGAIN)
				error = 0;
			if (error)
				saved_error = error;
		};

	/*
	 * Now adjust the refcount and take appropriate actions.
	 */
	if (--(name)->refcnt <= 0) {
		error = svcm_name_destroy(name, vp, cred);
	} else {
		error = svcm_unlock(name->lock);
	};

	return error ? error : saved_error;
}

/*-
 *--------------------------------------------------------------------------
 *
 * svcm_name_destroy -- called when all refcnts to a name are gone.
 *
 * DESCRIPTION:
 *	Cleans up after a name.
 *	Assumes that all callbacks on that name have already been 
 *	relinquished.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
int
svcm_name_destroy (name, vp, cred)
	struct svcm_name *name;		/* IN: locked name */
	struct vnode *vp;		/* IN: vp no longer needing refcnt */
	struct ucred *cred;		/* IN */
{
#ifdef SANITY_CHECKS
	if (!svcm_islocked(name->lock))
		panic("svcm_name_destroy: name is unlocked.");
#endif /* SANITY_CHECKS */

	/*
	 * Set death flag to catch death/re-birth race.
	 * We can't remove it from the hash chains yet because
	 * otherwise we might cache alias any cached resources
	 * beween a dieing and a recreated node.
	 */
	name->flags |= SVCM_NAME_DIEING;

	/*
	 * Now remove name from hash chains.
	 * Past this point, no other process can re-find the name.
	 * This is ok because we've just assured ourselves that
	 * we hold no important data.
	 */
	svcm_unhash_name(name);
	/*
	 * Wake up people waiting for our doom.
	 * They won't find us when they recover.
	 */
	if (name->flags & SVCM_NAME_REBORN) {
		/* printf ("Waking on name %x\n", u.u_procp->p_pid, name); */
		SVCM_STATS_INC(name_rebirth_wakes);
		wakeup((caddr_t)name);
	};

	/*
	 * Now, locking.
	 * Release and free our lock.
	 * NEEDSWORK: Race condition on lock users and freed up lock.
	 */
	svcm_unlock(name->lock);
	svcm_lock_rele(name->lock);

	/*
	 * Zero a few last things.
	 */
	name->refcnt = 0;
	name->flags = 0;

	/*
	 * And free ourselves.
	 */
	SVCM_STATS_INC(name_destroy);
	svcm_name_free(name);
	
	return 0;
}

#endif SVCM_TWO
