
/*
 * mmap_torture
 *
 * Run with:
 *     ./mmap_torture -vv /usr/tmp/s1 /mnt/{a,b}/d1
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <stdio.h>
#include <sys/param.h>
#include <i405/i405_syscalls.h>
#include <errno.h>
#include <ficus/fix_sunos_headers.h>

extern char *optarg;
extern int optind, opterr;



struct log {
	int testlayer;
	int testoff;
	int testval;
	int page;
};
int logend = 0;
struct log *log = NULL;

int verbose = 0;
int delay = 0;
#define MAXLAYERS 100	
char *testname[MAXLAYERS];
int testfd[MAXLAYERS];
u_char *goodaddr;
u_char *testaddr[MAXLAYERS];


void
usage()
{
	static char *message[] = {
		"usage: mmap_torture [-vsr] [-c count] [-S seed] [-l length] SingleLayer Layer1 Layer2 ...\n",
		"\tMemory maps SingleLayer and Layer1..n and then writes into count random\n",
		"\taddresses in SingleLayer and a random layer, checking for\n",
		"\tinconsistencies.\n",
		"\tOptions:\n",
		"\t\t-s sequential mode.\n",
		"\t\t-W read-only mode (no writes).\n",
		"\t\t-v verbose mode.\n",
		"\t\t-r rounding mode.\n",
		"\t\t-d debug mode.\n",
		"\t\t-c count of random things to try.  Defaults to 1000.\n",
		"\t\t-S Random number seed.  Default to zero.\n",
		"\t\t-l File length.  Default to 100k.\n",
		"\t\t-D Delay between tests.\n",
		"\t\t-P pre-touch all pages, then trap kgdb.\n",
		NULL,
	};
	char **p = message;
	while (*p)
		fprintf (stderr, "%s", *p++);
	exit (1);
}


void
dumplog(offset)
	int offset;
{
	int i;
	int ourpage = offset / PAGESIZE;

	for (i = 0; i < logend; i++) {
		if (log[i].page != ourpage)
			continue;
		printf ("%c %8x %8x %4x %2d\n",
		       (log[i].testoff == offset ? '*' : ' '),
		       log[i].page,
		       log[i].testoff,
		       log[i].testval,
		       log[i].testlayer);
	};
	fflush(stdout);
}

void
enter_kernel(name)
	char *name;
{
	/*
	 * Give chars time to be echoed.
	 */
	sleep(delay);
#if 1
	/*
	 * Do a random sys call so we can get the kernel debugger control.
	 * This will fail.
	 */
	syscall(SYS_vnode_type, name);
#else
	/*
	 * Do a random sys call so we can get the kernel debugger control.
	 * This will fail.
	 * (Put a trap on null_rename.)
	 */
	if (name)
		rename(name, "/");
#endif
}



int
openmap(fn, size, fdp, addrp, create_required, madvice)
	char *fn;
	int size;
	int *fdp;
	caddr_t *addrp;
	int create_required;
	int madvice;
{
	int fd;
	struct stat sb;
	caddr_t raddr, addr, mmap();
	unsigned char ff = 0xff;
	char tmps[256];
	int error;
	extern int errno;

	if (create_required) {
		error = stat(fn, &sb);
		if (error == 0 || (error == -1 && errno != ENOENT)) {
			fprintf (stderr, "file %s already exists\n", fn);
			if (error)
				perror ("stat");
			exit (1);
		};
	} else {
		if (-1 == stat(fn, &sb)) {
			sprintf (tmps, "stat: error on file %s", fn);
			perror (tmps);
			exit (1);
		};
	};
	
	if (-1==(fd = open(fn, O_RDWR|O_CREAT, 0666))) {
		sprintf (tmps, "open: %s", fn);
		perror (tmps);
		exit(1);
	};
	if (-1==fstat (fd, &sb)) {
		perror ("fstat");
		exit (1);
	};
	/*
	 * Make the file the right size.
	 * We don't use ftruncate because in SunOS this
	 * pre-allocates the data pages.
	 */
	if (sb.st_size < size) {
		if (size-1!=lseek(fd,size-1, SEEK_SET))
			perror("seek");
		if (1!=write (fd, &ff, 1))
			perror("write");
	};
	raddr = addr = 0;
	if (-1==(int)(raddr = mmap(addr, size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0))) {
		perror ("mmap");
		exit(1);
	};
	if (madvice)
		if (-1 == madvise (raddr, size, madvice)) {
			perror ("madvise");
			exit (1);
		};

	*fdp = fd;
	*addrp = raddr;

	if (verbose)
		printf ("open of <%s> is %d\n", fn, fd);
	
	return 0;
}

void
pre_page(addr, size)
	u_char *addr;
	int size;
{
	int i;
	u_char oldval;

	for (i=0; i<size; i+=PAGESIZE) {
		oldval = addr[i];
		addr[i] = 1;
		addr[i] = oldval;
	};
	enter_kernel();
}

int
main(argc, argv)
	int argc;
	char **argv;
{
	int count = 1000, size = 100*1024, seed = 0;
	int goodfd;
	int i = 1;
	int testlayer, testoff, testval, testpage;
	int layercount = 0;
	int sequential = 0;
	int rounding = 0;
	int debug = 0;
	int c;
	int problems = 0;
	int read_only_mode = 0;
	int pre_page_mode = 0;
	int create_required;
	int goodval, pre_testval, post_testval;

	while ((c = getopt (argc, argv, "c:dDl:PrsS:vW?")) != -1)
		switch (c) {
		case 's':
			sequential++;
			break;
		case 'r':
			rounding++;
			break;
		case 'v':
			verbose++;
			break;
		case 'd':
			debug++;
			break;
		case 'D':
			delay++;
			break;
		case 'c':
			count = atoi(optarg);
			break;
		case 'S':
			seed = atoi(optarg);
			break;
		case 'l':
			size = atoi(optarg);
			break;
		case 'W':
			read_only_mode++;
			break;
		case 'P':
			pre_page_mode++;
			break;
		default:
			usage();
		};
	if (argc - optind < 1)
		usage();

	if (verbose)
		printf ("pid %d\n", getpid());

	/*
	 * Set stuff up.
	 */
	srandom (seed);
	openmap(argv[optind++], size, &goodfd, &goodaddr, 1, (sequential ? MADV_RANDOM : 0));
	create_required = 1;
	for (layercount = 0; optind < argc; layercount++, optind++) {
		testname[layercount] = argv[optind];
		openmap(testname[layercount], size, &(testfd[layercount]), &(testaddr[layercount]), create_required, (sequential ? MADV_RANDOM : 0));
		create_required = 0;
	};
	if (verbose)
		printf ("layercount = %d\n", layercount);
	if (layercount == 0) {
		/*
		 * No other layer => nothing to verify.
		 * Fake up a "test layer" that's really the same as the "good
		 * layer".
		 */
		layercount = 1;
		testname[0] = "internal";
		testfd[0] = goodfd;
		testaddr[0] = goodaddr;
	};
	if ((log = (struct log *)malloc(sizeof(struct log) * count)) == NULL) {
		fprintf (stderr, "cannot allocate log (out of memory)");
		exit (1);
	};

	if (pre_page_mode)
		pre_page(testaddr[layercount-1], size);

	testval = 0;
	for (i = 0; i < count; i++) {
		testlayer = random() % layercount;
		testoff = random() % size;
		if (sequential)
			testoff = i % size;
		testpage = testoff / PAGESIZE;
		if (rounding)
			testoff = testpage * PAGESIZE;
		testval = (testval+1) % 256;

		log[i].testlayer = testlayer;
		log[i].testoff = testoff;
		log[i].testval = testval;
		log[i].page = testpage;
		logend = i + 1;

		if (verbose) {
			printf ("trial: offset %x (page %x) with layer %d value %x\n",
				testoff, testpage, testlayer, testval);
			/* In case we crash, make sure our data is written. */
			if (verbose > 1) {
				fflush(stdout);
			};
			if (delay)
				sleep(delay);
		};

		/* First pre-read the thing */
		goodval = goodaddr[testoff];
		pre_testval = testaddr[testlayer][testoff];
		if (goodval != pre_testval) {
			char tmps[128];
			sprintf (tmps, "%s: offset %x pre-check fails for layer %d, good=%x, test=%x",
				testname[testlayer],
				testoff, 
				testlayer,
				goodval,
				pre_testval);
			printf ("%s\n", tmps);
			if (verbose) {
				dumplog(testoff);
				enter_kernel(testname[testlayer]);
			};
			problems++;
		};
		/* Write the new value */
		if (!read_only_mode) {
			goodaddr[testoff] = testval;
			testaddr[testlayer][testoff] = testval;
		};
		/* Read it back to check it. */
		goodval = goodaddr[testoff];
		post_testval = testaddr[testlayer][testoff];
		if (goodval != post_testval) {
			printf ("offset %x read-back fails for layer %d, good=%x, test=%x\n",
				testoff, testlayer,
				goodval, post_testval);
			problems++;
		};
	};
	exit (problems ? 1 : 0);
}

