
/*-
 * i405_init.c
 *
 *	@(#)i405_init.c	2.32 (UCLA Ficus Project) 1/9/95
 *
 * Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * This has all the configuration stuff for the new interface.
 * It compiles in and out of the kernel, handling things either way.
 * 
 * Define "BACKVINT" to fill in the first ops with ops that will
 * translate the old interface to the new.
 */


/*
 * STANDALONE_OOK must be undefed because otherwise
 * we get the wrong definition of VOFFSET 
 * (and hence compiler warnings )when compiling
 * OOK with gcc.
 */
#ifdef STANDALONE_OOK
#undef STANDALONE_OOK
#endif


#include <sys/param.h>
#include <sys/time.h>
#include <sys/vnode.h>
#include <sys/vfs.h>
#include <i405/i405.h>
#include <sys/errno.h>
#include <sys/user.h>
#include <ficus/cdefs.h>

#include <i405/i405_subr.h>
#ifdef BACKVINT
#include <i405/i405_backvint.h>
#endif


#ifdef NOTDEF
#define DODEBUG(A) A
#else
#define DODEBUG(A)
#endif


#if defined(BACKVINT) && defined(OUTOFKERNEL)
Backwards compatibility is not supported in the out-of-kernel interface.
#endif




extern struct vnodeopv_desc *i405_opv_descs[];
				/* a list of lists of vnodeops defns */
extern struct vnodeop_desc *i405_op_descs[];
				/* and the operations they perform */
/*
 * This code doesn't work if the defn is **vnodop_defns with cc.
 * The problem is because of the compiler sometimes putting in an
 * extra level of indirection for arrays.  It's an interesting
 * "feature" of C.
 */
extern int num_operations;

#ifndef __PFI_defined
typedef int (*PFI)();   /* the standard Pointer to a Function returning an Int */
#define __PFI_defined
#endif



/*-
 *--------------------------------------------------------------------------
 *
 * i405_opv_set_implementation -- set an implementation for an op in an opv
 *
 * DESCRIPTION:
 *	An exported interrface (for Ted's OO stuff) to
 *	set an implementation for an op in an opv.
 *
 * RETURN VALUES:
 *      Panics on errors.
 *
 *--------------------------------------------------------------------------
 */
void
i405_opv_set_implementation(opv, descp, impl)
	PFI *opv;
	struct vnodeop_desc *descp;
	PFI impl;
{
	/*
	 * Sanity check:  is this operation listed
	 * in the list of operations?  We check this
	 * by seeing if its offest is zero.  Since
	 * the default routine should always be listed
	 * first, it should be the only one with a zero
	 * offset.  Any other operation with a zero
	 * offset is probably not listed in
	 * i405_op_descs,
	 * and so is probably an error.
	 *
	 * A panic here means the layer programmer
	 * has committed the all-too common bug
	 * of adding a new operation to the layer's
	 * list of supported vnode operations but
	 * not adding the operation to the system-wide
	 * list of supported operations.
	 */
	if (descp->vdesc_offset == 0 && descp != VDESC(nvop_default)) {
		printf ("i405_opv_set_implementation: name %s is not listed in i405_op_descs.\n",
				descp->vdesc_name);
		panic("i405_opv_set_implementation: attempt to configure an uninitialized operation.");
	};
	opv[descp->vdesc_offset] = impl;
}


/*-
 *--------------------------------------------------------------------------
 *
 * i405_opv_populate -- add operations to an ops vector
 *
 * DESCRIPTION:
 *	Add operations to an ops vector, allocating it if necessary.
 *	
 *	An undocumented feature of this approach to defining
 *	operations is that there can be multiple entries in
 *	i405_opv_descs for the same operations vector.
 *	This allows third parties to extend the set of operations
 *	supported by another layer in a binary compatibile way.
 *	For example, assume that NFS needed to be modified to support
 *	Ficus.  NFS has an entry (probably nfs_vnodeop_opv_desc)
 *	declaring all the operations NFS supports by default.
 *	Ficus could add another entry (ficus_nfs_vnodeop_opv_entensions)
 *	listing those new operations Ficus adds to NFS, all without
 *	modifying the NFS code.  (Of couse, the OTW NFS protocol
 *	still needs to be munged, but that's a(whole)nother story.)
 *	This is a feature.
 *
 * SEE ALSO:
 *	i405_opv_default.
 *
 * RETURN VALUES:
 *      None.
 *
 *--------------------------------------------------------------------------
 */
static void
i405_opv_populate(opv_desc_vector_p, opv_desc_ops)
	PFI **opv_desc_vector_p;
	struct vnodeopv_entry_desc *opv_desc_ops;   /* null terminated list */
{
	int j;
	int (**opv_desc_vector)();
	struct vnodeopv_entry_desc *opve_descp;

	/*
	 * Allocate the vector, if necessary.
	 */
	if (*opv_desc_vector_p == NULL) {
		*opv_desc_vector_p = (PFI*)
			kmem_zalloc(num_operations*sizeof(PFI));
#ifdef BACKVINT
		bcopy(&oldtonew_vnodeops, *opv_desc_vector_p,
		      num_back_operations*sizeof(PFI));
#endif
	};


	/*
	 * Now add the new operations.
	 */
	opv_desc_vector = *opv_desc_vector_p;
	for (opve_descp = &opv_desc_ops[0], j = 0;
	     opve_descp->opve_op;
	     opve_descp++, j++) {
		i405_opv_set_implementation(opv_desc_vector, opve_descp->opve_op, opve_descp->opve_impl);
	};
}


/*-
 *--------------------------------------------------------------------------
 *
 * i405_opv_default -- fill in the default operations of an ops vector
 *
 * DESCRIPTION:
 *	Replace all unfilled ops vector entries with the default
 *	routine.
 *
 * SEE ALSO:
 *	i405_opv_populate.
 *
 * RETURN VALUES:
 *      None.
 *
 *--------------------------------------------------------------------------
 */
static void
i405_opv_default(opv_desc_vector_p)
	PFI **opv_desc_vector_p;
{
	int k;
	int (**opv_desc_vector)() = *opv_desc_vector_p;

	if (!opv_desc_vector_p)
		panic("i405_opv_default: unallocated ops vector.");
	opv_desc_vector = *opv_desc_vector_p;

	/*
	 * Force every operations vector to have a default routine.
	 */
	if (opv_desc_vector[VOFFSET(nvop_default)] == NULL) {
		panic("i405_opv_default: operation vector without default routine.");
	};

	for (k = 0; k < num_operations; k++)
		if (opv_desc_vector[k] == NULL)
			opv_desc_vector[k] = 
				opv_desc_vector[VOFFSET(nvop_default)];
}



#if 0   /* defined(JOHNH) */
#define I405_BOOT_DEBUG 1
#else 
#define I405_BOOT_DEBUG 0
#endif 
int i405_boot_debug = I405_BOOT_DEBUG;


/*-
 *--------------------------------------------------------------------------
 *
 * i405_init -- init all vnodeop vectors
 *
 *
 * DESCRIPTION:
 *	Init all vnodeop vectors, figuring out how big they should
 *	be.  We call i405_opv_populate to add to each operation
 *	vector, then call i405_opv_default to fill in the bypass
 *	routine when done.
 *
 * 	This routine should be called from the kernel's main
 *	before any vnodes are used.  I put it just before
 *	quota initialization.
 *
 * SEE ALSO:
 *	i405_opv_init
 *	
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
void
i405_init()
{
	int i;

	DODEBUG(printf ("i405_init.\n"));

	/*
	 * Sanity-check some assertions.
	 */
#ifdef SVCM_SFO
#ifdef SVCM_PAGEFLIP
	panic("i405_init: SVCM_SFO and SVCM_PAGEFLIP are both defined.");
#endif /* SVCM_PAGEFLIP */
#ifdef UFS_USE_SVCM_DNLC
	panic("i405_init: SVCM_SFO and UFS_USE_SVCM_DNLC are both defined.");
#endif /* UFS_USE_SVCM_DNLC */
#endif /* SVCM_SFO */

	/*
	 * Set all vnode vectors to a well known value.
	 */
	for (i=0; i405_opv_descs[i]; i++)
		*(i405_opv_descs[i]->opv_desc_vector_p) = NULL;

#ifdef BACKVINT
	/*
	 * Leave space at the beginning for old compatibility functions.
	 */
#ifdef I_FWL
	/* We allocate backwards ops with the oldtonew FWL. */
#else /* ! I_FWL */
	num_operations = num_back_operations;
	DODEBUG(printf ("ops 0-%d reserved for compatibility\n", num_operations-1));
#endif /* I_FWL */
#else
	num_operations = 0;
#endif


	/*
	 * Figure out how many ops there are by counting the table,
	 * and assign each its offset.
	 */
	for (i=0; i405_op_descs[i]; i++) {
		i405_op_descs[i]->vdesc_offset = num_operations;
		num_operations++;
	};

	DODEBUG(printf ("num_operations=%d\n", num_operations));

	/*
	 * First populate all vectors, then go back and
	 * fill in the defaults.  These are separate steps
	 * so that multiple people can take a whack
	 * at the ops vectors before defaulting is done.
	 */
	for (i = 0; i405_opv_descs[i]; i++) {
		i405_opv_populate(i405_opv_descs[i]->opv_desc_vector_p,
				  i405_opv_descs[i]->opv_desc_ops);
	};
	for (i = 0; i405_opv_descs[i]; i++) {
		i405_opv_default(i405_opv_descs[i]->opv_desc_vector_p);
	};

#ifdef I_FWL
#ifdef BACKVINT
	/*
	 * Sanity check the backwards compatibilty stuff.
	 */
	oldtonew_sanity_check();
#endif /* BACKVINT */
#endif /* I_FWL */

#if defined(GPROF) && defined(FICUS)
	ficus_patch_opvectors_for_gprof();
#endif
	i405_init_dead_vp();

	DODEBUG(printf ("i405 init done.\n"));
}


