
/*
 * i405_syscalls.c
 *
 *	@(#)i405_syscalls.c	2.25 (UCLA Ficus Project) 3/6/95
 *
 *
 * Copyright (C) 1992 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * User system calls to the new vnode interface.
 *
 * The uap's below aren't the ones that come directly from the
 * system call.  They're just pointers into user space
 * of the user's vnode args struct.
 *
 * Vnodes are given to the user level as file descriptiors.
 * This means that when a user program crashes, its vnodes
 * will be released.
 *
 * To support the null layer out-of-kernel (i.e., under nfs),
 * we need to be able to to get the vfs structure for
 * a vnode, and to do a vget on that vfs.  As a temporary
 * patch, these are new system calls.  The vfs is given
 * to the user as a pointer, but list of current mounted
 * vfs's is always checked to make sure the pointer is valid.
 *
 */
 
 
#include <sys/param.h>
#ifdef OUTOFKERNEL
#include "myuser.h"
#else
#include <sys/user.h>
#endif
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/vfs.h>
#include <sys/vnode.h>
#include <i405/i405.h>
#include <rpc/types.h>
#include <rpc/xdr.h>
#include <sys/syslog.h>

#include <ficus/cdefs.h>
#ifdef BACKVINT
#include <i405/i405_backvint.h>
#endif
#ifdef INTGEN_TWO
#include <i405/i405_subr.h>
#include <i405/i405_utok_fdvp.h>
#endif
#ifdef NEWOFS
#include <newofs/newo_interface.h>
#endif NEWOFS

#if 0
#define DODEBUG(A) A
#else
#define DODEBUG(A)
#endif


extern int num_operations;
#ifdef BACKVINT
extern int num_back_operations;
#endif
extern struct vnodeop_desc *vnodeop_descs[];



/*
 *--------------------------------------------------------------------------
 *
 * user_vnodeop -- the syscall entry for user-level vnodeops
 *
 *
 * DESCRIPTION:
 *	Invoke a user-requested vnode operation on a user-level vnode
 *	(provided as a file descriptor).  The basic plan is simple:
 *      Bring all arguments into kernel space with XDR, call the op,
 *      then use XDR to push the results back out.
 *
 * RETURN VALUES:
 *      Returns error code in u.u_error (Sun's convention).
 *
 *--------------------------------------------------------------------------
 */
void
user_vnodeop(uap)
	/*NOPROTO*/
        struct a {
                int uvp;		/* IN: user-level vnode that op
					 * will be invoked on. */
		struct nca_uuid *uuidp; /* IN: uuid identifying op */
                struct nvop_generic_args **app;
					/* INOUT: Ptr to op arguments and
					 * results. */
        } *uap;
{
	struct nca_uuid uuid;
        struct vnodeop_desc *vdescp;
	int offset;
	struct ucred *tmpcr = NULL;
	struct vnode *vp = NULL;
	struct nvop_generic_args *args = NULL;
	int error = 0, alternate_error;
		
	/*
	 * First get and map uuid to operation,
	 * bailing on unknown ops.
	 */
	if (error = copyin(uap->uuidp, &uuid, sizeof(struct nca_uuid))) {
		u.u_error = error;
		return;
	};
	if (NULL == (vdescp = i405_uuid_to_desc(&uuid))) {
		u.u_error = EINVAL;
		return;
	};
	offset = vdescp->vdesc_offset;

	/*
	 * Inactive requires special help.  Handle it here.
	 */
	if (vdescp == &nvop_inactive_desc) {
		u.u_error = user_inactive(uap->uvp, *(uap->app));
		return;
	};

	/*
	 * First things first:  bring the args into kernel space.
	 * (Note that after this point the only correct way
	 * out of the function is via "out" so that we can 
	 * free whatever we've alloc'ed.)
	 */
	DODEBUG(printf("Calling i405_utok_reference on the way in.\n"));
	if (error = i405_utok_reference (uap->app, &args,
					 vdescp->vdesc_datatype, TRANSPORT_UCLA_UTOK, WHY_IN)) {
		log (LOG_WARNING, "user_vnodeop: i405_utok_reference errors with %d on copyin.\n", error);
		goto out;
	};
	DODEBUG(printf("Back from i405_utok_reference\n"));

	/*
	 * Now that we've got the operation done,
	 * let's go fill in the vdesc pointer.
	 * (Yes, it's a little odd that we didn't do
	 * this earlier, but we couldn't because the
	 * storage wasn't allocated.)
	 */
	if (!args)
		panic("user_vnodeop: args not allocated but error not returned.");
	args->a_desc = vdescp;

	/*
	 * Find the vnode in the args.
	 * Special case VOP_OPEN with its obnoxious vpp.
	 */
	if (vdescp->vdesc_flags & VDESC_INVPP) {
		if (vdescp != VDESC(nvop_open))
			panic ("user_vnodeop: invpp set but op is not vop_open.");
		vp = *(((struct nvop_open_args *)args)->a_vpp);
	} else {
		if (vdescp->vdesc_vp_offsets[0] == VDESC_NO_OFFSET)  /* sanity */
			panic ("user_vnodeop: operation with no vnode not allowed.");
		vp = *VOPARG_OFFSETTO(struct vnode**, vdescp->vdesc_vp_offsets[0], args);
	};
	if (!vp) {
		DODEBUG(printf ("user_vnodeop: vp is null\n"));
		error = EINVAL;
		goto out;
	};
	DODEBUG(printf ("calling, vp=%x, err=%d \n", vp, u.u_error));

	/*
	 * What a hack.  Some Sun file systems (like UFS)
	 * take the uid out of u.u_cred, not the cred field
	 * passed to the function.
	 *    So to make this work, we make them both the same.
	 *    No self-respecting, modular-claiming layer should
	 * ever do this.
	 */
	if (vdescp->vdesc_cred_offset != VDESC_NO_OFFSET) {
		tmpcr = u.u_cred;
		u.u_cred = *(struct ucred**)
			(((char*)args)+vdescp->vdesc_cred_offset);
	};

#ifdef BACKVINT
	/*
	 * Bail out if we're not on a i405-enabled file system.
	 * This grossness is required because we don't convert
	 * all file system types.
	 *
	 * To demonstrate the bug this avoids, try getfattrs on different
	 * file systems.  On UFS (which is converted) it will return
	 * EOPNOTSUP, but on /tmp it will crash the machine
	 * because tmpfs isn't converted.
	 */
	if (vp->v_op->vn_close != oldtonew_vnodeops.vn_close) {
		log (LOG_DEBUG, "user_vnodeop: new operation attempted on non-converted file system type.\n");
		error = EOPNOTSUPP;
		goto skipit;   /* don't do if/else because of #ifdef */
	};

#endif

	/*
	 * Call the routine.
	 */
	error = VCALL(vp, offset, args);
 skipit:

	/*
	 * Undo the cred grossness.
	 */
	if (tmpcr) {
		u.u_cred = tmpcr;
		tmpcr = NULL;
	};

	/*
	 * Now send out the results.
	 */
	DODEBUG(printf ("calling i405_utok_reference to put the results.\n"));
	if (!error) {
		if (error = i405_utok_reference (uap->app, &args,
				vdescp->vdesc_datatype,
				TRANSPORT_UCLA_UTOK, WHY_OUT)) {
			log (LOG_WARNING, "user_vnodeop: i405_utok_reference returned error %d on copyout.\n", error);
			goto out;
		};
	};

out:
	/*
	 * From here we'll free our memory and exit.
	 * We must be careful to preserve the value stored
	 * in "error" if it's non-zero.
	 */

	/*
	 * Free our memory.
	 */
	DODEBUG(printf ("calling i405_utok_reference to free the args.\n"));
	if (alternate_error = i405_utok_reference (NULL, &args,
				vdescp->vdesc_datatype,
	        		TRANSPORT_UCLA_UTOK, WHY_FREE)) {
		log (LOG_WARNING, "user_vnopdeop: i405_utok_reference returns error %d on free.\n", alternate_error);
		/* goto out :-) */
	};
	if (!error)
		error = alternate_error;

	DODEBUG(printf ("exiting, error=%d\n", error));

	u.u_error = error;
}



/*
 *--------------------------------------------------------------------------
 *
 * vnode_to_vfs -- syscall to convert user-level vnode to vfs pointer
 *
 * DESCRIPTION:
 *	Given a user vnode, get its vfsp and *vfsp.
 *	This routine works in concert with vnode_to_fid and
 *      fid_to_vnode to provide a way to map between fid and
 *	vnodes at the user-level.
 *
 *	This operation should be used only by the utok layer;
 *	it's not intended to be a generally availabe service.
 *	HOWEVER, the function vfs_to_vnode will do the reverse
 *	translation.
 *
 * SEE ALSO:
 *	fid_to_vnode, vfs_to_vnode
 *
 * RETURN VALUES:
 *      Returns error code in u.u_error (Sun's convention).
 *
 *--------------------------------------------------------------------------
 */
void
vnode_to_vfs(uap)
	/*NOPROTO*/
	struct a {
		int uvp;		/* IN: user-level vnode */
		struct vfs **vfsp;	/* OUT: vfs structure, NULL if none */
		struct vfs *vfs;	/* OUT: vfs pointer or NULL if none */
	} *uap;
{
	struct vnode *vp;

#ifdef INTGEN_TWO
	u.u_error = fdtovp(uap->uvp, &vp);
#else
	u.u_error = 0;
	if (NULL == (vp = fdtovp(uap->uvp)))
		u.u_error = EINVAL;
#endif
	if (!u.u_error && uap->vfsp)
		u.u_error = copyout(&vp->v_vfsp,uap->vfsp,sizeof(struct vfs*));
	if (!u.u_error && uap->vfs)
		u.u_error = copyout(vp->v_vfsp,uap->vfs,sizeof(struct vfs));
}



/*-
 *--------------------------------------------------------------------------
 *
 * vfs_to_vnode -- map a vfsp to its root vnode
 *
 * DESCRIPTION:
 *	Map a vfsp (from vnode_to_vfs) to its root vnode.
 *
 * SEE ALSO:
 *	vnode_to_vfs.
 *
 * RETURN VALUES:
 *       Returns error code in u.u_error (Sun's convention).
 *
 *--------------------------------------------------------------------------
 */
void
vfs_to_vnode(uap)
	/*NOPROTO*/
	struct a {
		struct vfs *vfsp;   /* IN: vfsp */
		int *uvp;	    /* OUT: user-level vnode for root of vfsp */
	} *uap;
{
	int error;
	struct vfs *ptr;
	struct vnode *vp = NULL;
	int fd;
	
	for (ptr = rootvfs; ptr; ptr = ptr->vfs_next)
		if (ptr == uap->vfsp)
			break;
	if (!ptr)
		error = ESTALE;

	if (error = VFS_ROOT(uap->vfsp, &vp))
		goto out;

	if (error = vptofd(vp, &fd))
		goto out;

	error = copyout(&fd, uap->uvp, sizeof(int));

out:
	if (error) {
		if (vp)
			VN_RELE(vp);
	};
	/* return value to user */
	u.u_error = error;
}



/*
 *--------------------------------------------------------------------------
 *
 * vnode_to_fid -- syscall to convert user-level vnode to a fid
 *
 *
 * DESCRIPTION:
 *	Given a user-level vnode, return its fid to the user.
 *	This is required to be a special syscall since VOP_FID
 * 	returns a variable length of data, in voilation of the
 *	user-to-kernel transport requirements.
 *
 *	A preferred way to do this is to use VOP_NEWFID
 *	to get the fid and the newfid_to_vnode to map back.
 *	You then use VOP_NEWFID_TO_VP to map back.
 *	Unfortunately this procedure only works in layers supporting
 *	VOP_NEWFID.
 *
 * SEE ALSO:
 *	fid_to_vnode, vnode_to_vfs
 *
 * RETURN VALUES:
 *      Returns error code in u.u_error (Sun's convention).
 *
 *--------------------------------------------------------------------------
 */
void
vnode_to_fid (uap)
	/*NOPROTO*/
	struct a {
		int uvp;		/* IN: user-level vnode */
		struct fid *fid;	/* OUT: fid (user should alloc max size) */
	} *uap;
{
	struct vnode *vp;
#ifndef INTGEN_TWO
	struct vnode *fdtovp();
#endif
	struct fid *fid;
	USES_NVOP_FID;
	
#ifdef INTGEN_TWO
	u.u_error = fdtovp(uap->uvp, &vp);
#else
	u.u_error = 0;
	if (NULL == (vp = fdtovp(uap->uvp)))
		u.u_error = EINVAL;
#endif
	if (!u.u_error)
		u.u_error = NVOP_FID(vp, &fid);
#if 0   /* for debugging */
	{       int i;
		printf ("vnode_to_fid: len=%d, ", fid->fid_len);
		for (i=0; i<10; i++)
			printf ("%2x", fid->fid_data[i]);
		printf ("\n");
	}
#endif
	if (!u.u_error)
		u.u_error = copyout((caddr_t)fid,
				    (caddr_t)uap->fid,
				    fid->fid_len+sizeof(short));
}


/*
 *--------------------------------------------------------------------------
 *
 * fid_to_vnode -- syscall to convert vfs/fid to vnode to a user-level vnode
 *
 *
 * DESCRIPTION:
 *	Given a vfsp and a fid, return the designated vnode.
 *
 * SEE ALSO:
 *	fid_to_vnode, vnode_to_vfs
 *
 * RETURN VALUES:
 *      Returns error code in u.u_error (Sun's convention).
 *
 *--------------------------------------------------------------------------
 */
void
fid_to_vnode (uap)
	/*NOPROTO*/
	struct a {
		struct vfs *vfsp;	/* IN: vfsp from vnode_to_vfs */
		struct fid *fidp;	/* IN: fid from vnode_to_fid */
		int *uvp;		/* OUT: user-level vnode */
	} *uap;
{
	struct vfs *vfsp;
	struct vnode *vp;
	struct fid fid;
	int fd;

	/*
	 * First get and identify the vfs.
	 */
	u.u_error = 0;
	for (vfsp=rootvfs; vfsp; vfsp=vfsp->vfs_next)
		if (vfsp==uap->vfsp)
			break;
	if (!vfsp)
		u.u_error = ESTALE;

	/*
	 * Now do the fid.
	 *
	 * This two part copy of the fid madness is necessary because
	 * it's conceivable that the user has only allocated part
	 * of the fid, so copying sizeof(struct fid) could cause
	 * a memory fault.   -JSH 8-Feb-91
	 */
	if (!u.u_error)
		u.u_error = copyin(uap->fidp, &fid, sizeof(short));
	if (!u.u_error && fid.fid_len)
		if (/*fid.fid_len>=0 &&*/ fid.fid_len<MAXFIDSZ)  /* fid_len is unsigned, so always > 0*/
			u.u_error = copyin(uap->fidp, &fid, sizeof(short)+fid.fid_len);
		else u.u_error = EINVAL;

	/*
	 * VGET is the whole point of this routine.
	 * Get the vnode and then convert it to userism.
	 */
	if (!u.u_error)
		u.u_error = VFS_VGET(vfsp, &vp, &fid);

	/*
	 * ufs_vget doesn't return an error from vfs_vget,
	 * it just returns a null vp.  Grrrr.
	 * Turn this failure into an error code here.
	 */
	if (vp == NULL) {
		u.u_error = ESTALE;
	};

#ifdef INTGEN_TWO
	if (!u.u_error)
		u.u_error = vptofd(vp, &fd);
#else
	if (!u.u_error)
		fd = vptofd(vp);
#endif

	if (!u.u_error)
		u.u_error = copyout(&fd, uap->uvp, sizeof(int));
}





/*
 *--------------------------------------------------------------------------
 *
 * ook_lookupname -- lookupname (pathname translation) on user-level vnodes
 *
 *
 * DESCRIPTION:
 *	This routine does pathname translation to and from user-level vnodes.
 *	It exists to provide exact lookupname semantics to the user.
 *	(It's difficult and expensive to correctly support "followlink"
 *	with strictly existing syscalls, and silly and expensive to re-implement
 *	lookupname at the user-level using vnode operations.)
 *
 * RETURN VALUES:
 *      Returns error code in u.u_error (Sun's convention).
 *
 *--------------------------------------------------------------------------
 */
void
ook_lookupname (uap)
	/*NOPROTO*/
	struct a {
		char *name;
		enum symfollow followlink;
		struct vnode **dirvpp;
		struct vnode **compvpp;
	} *uap;
{
	struct vnode *dirvp = NULL, *compvp = NULL;
	int dirfd = -1, compfd = -1;
	int error = 0;

	DODEBUG(printf("ook_lookupname: name at %x, fl=%d, dirvpp=%x, compvpp=%x\n",
		       uap->name, uap->followlink, uap->dirvpp, uap->compvpp));

	/*
	 * Do the lookup.
	 */
	if (error = lookupname(uap->name, UIO_USERSPACE, uap->followlink,
				   ((uap->dirvpp==NULL) ? NULL : &dirvp), &compvp))
		goto badout;

	/*
	 * Map results to fds and return them.
	 */
	if (uap->dirvpp) {
#ifdef INTGEN_TWO
		if (error = vptofd(dirvp, &dirfd))
			goto badout;
#else /* INTGEN_TWO */
		dirfd = vptofd(dirvp);
#endif /* INTGEN_TWO */
		dirvp = NULL;   /* refcount now held by dirfd */
		if (error = copyout(&dirfd, uap->dirvpp, sizeof(int)))
			goto badout;
	}
#ifdef INTGEN_TWO
	if (error = vptofd(compvp, &compfd))
		goto badout;
#else /* INTGEN_TWO */
	compfd = vptofd(compvp);
#endif /* INTGEN_TWO */
	compvp = NULL;   /* refcount now held by compfd */
	if (error = copyout(&compfd, uap->compvpp, sizeof(int)))
		goto badout;

	u.u_error = 0;
	u.u_rval1 = 0;	/* vptofd sets this field and it has to be zero. */
	return;

 badout:
	if (dirfd != -1)
		cleanupfd(dirfd);
	if (compfd != -1)
		cleanupfd(compfd);
	if (compvp)
		VN_RELE(compvp);
	if (dirvp)
		VN_RELE(dirvp);
	u.u_error = error;
}



/*
 *--------------------------------------------------------------------------
 *
 * normalize_vnode -- check for duplicate ook vnodes
 *
 * DESCRIPTION:
 *	OOK, each vnode *reference* (as opposed to each vnode)
 *	is represented by an fd.  User-level layers want to be
 *	able to unify these references.  Normally they'd do
 *	this by VOP_CMP'ing a returned fd against all open
 *	fds.  This is both expensive and painful to do at user-level,
 *	so we do it here.
 *     
 *	This function takes an fd as input and returns an fd.
 *	If the input fd is already represented in the processes
 *	fd list, we return its friend (we expect the caller to release
 *	one of the two).  Otherwise we just return it.
 *
 * RETURN VALUES:
 *      Returns error code in u.u_error (Sun's convention).
 *
 *--------------------------------------------------------------------------
 */
void
normalize_vnode(uap)
	/*NOPROTO*/
	struct a {
		int newfd;	/* IN: a (possibly) abnormal fdvp */
		int *oldfdp;	/* OUT: a normal vp */
	} *uap;
{
	int indx;
	struct file *oldfp, *newfp;
	int oldfd = -1;
	struct vnode *vp;

	if (NULL == (newfp = getf(uap->newfd))) {
		u.u_error = EBADF;
		return;
	};
	vp = (struct vnode*)newfp->f_data;

	/*
	 * Look for a duplicate.
	 */
	for (indx = 1; indx <= u.u_lastfile; indx++)
		if (indx != uap->newfd && (oldfp = u.u_ofile[indx]) != NULL)
			if ((oldfp->f_count > 0) &&
			    (oldfp->f_type == DTYPE_OURS) &&
			    ((struct vnode*)(oldfp->f_data) == vp)) {
				/*
				 * Duplicate exists.  Return old.
				 */
				oldfd = indx;
#if 0
				/*
				 * Now we expect the caller to 
				 * release one of the two.
				 */
				/*
				 * Call the close system call with our
				 * uap as the arguments.
				 *
				 * NEEDSWORK: this is a hacky way to invoke
				 * close from within the kernel, but the 
				 * alternative is to duplicate a bunch of
				 * Sun code here.
				 */
				close();
#endif
				break;
			};
	/*
	 * If no duplicate, just return the old one.
	 */
	if (oldfd == -1)
		oldfd = uap->newfd;
	u.u_error = copyout(&oldfd, uap->oldfdp, sizeof(int));
}


/*
 *--------------------------------------------------------------------------
 *
 * vnode_to_fd -- convert an OOK vnode into a file descriptor
 *
 * DESCRIPTION:
 *	Ted needs to convert an OOK vnode to a full file descriptor
 *	so he can do things like stat.
 *
 * RETURN VALUES:
 *      Returns error code in u.u_error (Sun's convention).
 *
 *--------------------------------------------------------------------------
 */
void
vnode_to_fd(uap)
	/*NOPROTO*/
	struct a {
		int fdvnode;	/* IN: a fd representing a vnode */
		int filemode;   /* IN: open(2)-style mode */
		int *fdfd;	/* OUT: a normal fd */
	} *uap;
{
	struct file *vnodefp, *fdfp;
	int fdindx;
	extern struct fileops vnodefops;
	struct vnode *vp;

	if (NULL == (vnodefp = getf(uap->fdvnode))) {
		u.u_error = EBADF;
		return;
	};
	/* Make sure we're converting the right thing. */
	if (vnodefp->f_type != DTYPE_OURS) {
		u.u_error = EINVAL;
		return;
	};
	vp = (struct vnode*)vnodefp->f_data;

	/*
	 * Allocate a new fd/file table entry.
	 */
	fdfp = falloc();
	if (fdfp == NULL)
		return;   /* error already left in u.u_error */
	fdindx = u.u_r.r_val1;		/* this is bullshit */

	/*
	 * Set up the new entry.
	 */
	fdfp->f_flag = (uap->filemode - FOPEN) & FMASK;
	fdfp->f_type = DTYPE_VNODE;
	VN_HOLD(vp);
	fdfp->f_data = (caddr_t)vp;
	fdfp->f_ops = &vnodefops;

	/* return the new fd */
	u.u_error = copyout(&fdindx, uap->fdfd, sizeof(int));
}
