
/*
 * i405_transport.c
 *
 *	@(#)i405_transport.c	2.3 (UCLA Ficus Project) 5/18/94
 *
 * Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <rpc/xdr.h>
#include <sys/vnode.h>
#include <i405/i405_transport.h>
#include <sys/errno.h>

/* ook_compat.h has too much baggage */
typedef int void_but_int;
extern void_but_int bzero __P((void *b, int length));
extern int kmem_free __P((caddr_t ptr, u_int nbytes));
extern void panic __P((char *s)) func_noreturn;
extern int printf __P((const char *, ...));


/*
 * i405_excessive_array_elements specifies
 * an overflow value.  Arrays passing in more
 * than this number of elements are assumed
 * to be in error.
 *
 * NEEDSWORK: Rather than a fixed value,
 * this should be computed relative to the
 * size of an individual element.
 */
int i405_excessive_array_elements = 64*1024;


/*
 * Each parameter is context information:
 *   there - Information about "the other side".
 *		For XDR, this is a struct xdr *.
 *		For utok, this will be the remote pointer stack.
 *   here - Information about "this side".
 *		Where the data will go or already is.
 *   what_d - What the data is.  struct xport_datatype.
 *   how - What protocol we're using.  An offset into the protocol ops vector.
 *   why - Why we're doing this: one of alloc, free, in_encode, in_decode,
 * 		out_encode, out_decode.
 */
#ifdef NOTDEF
int
i405_datatype_interpret(there, here, what_d, how, why)
	caddr_t there, here;
	struct xport_datatype *what_d;
	int how;
	int why;
{
	return (*di->datatype->fns[xport_no])
		(xport_closure, data, di->instance_closure);
}
#endif


/*
 * Declare method functions for each transport type.
 */

PFI i405_indirect_fns[] = {
	i405_xdr_reference,
	i405_utok_reference,
	i405_alloc_yes,
	i405_xdr_reference,
};
PFI i405_implicit_array_fns[] = {
	i405_xdr_implicit_array,
	i405_utok_implicit_array,
	i405_alloc_yes,
	i405_xdr_implicit_array,
};
PFI i405_indirect_implicit_array_fns[] = {
	i405_xdr_indirect_implicit_array,
	i405_utok_indirect_implicit_array,
	i405_alloc_yes,
	i405_xdr_indirect_implicit_array,
};
PFI i405_fixed_array_fns[] = {
	i405_xdr_fixed_array,
	i405_utok_fixed_array,
	i405_alloc_yes,
	i405_xdr_fixed_array,
};
PFI i405_indirect_fixed_array_fns[] = {
	i405_xdr_indirect_fixed_array,
	i405_utok_indirect_fixed_array,
	i405_alloc_yes,
	i405_xdr_indirect_fixed_array,
};
PFI i405_indirect_variable_array_fns[] = {
	i405_xdr_indirect_variable_array,
	i405_utok_indirect_variable_array,
	i405_alloc_yes,
	i405_xdr_indirect_variable_array,
};



/*
 *--------------------------------------------------------------------------
 *
 * i405_einval -- a stub routine for the new XDR stuff
 *
 * DESCRIPTION:
 *	Just return an error EINVAL.
 *
 * RETURNS:
 *	The errno value EINVAL.
 *
 *--------------------------------------------------------------------------
 */
int
i405_einval()
{
	return EINVAL;
}

int
i405_noerror()
{
	return 0;
}



/*
 *--------------------------------------------------------------------------
 *
 * i405_free_array -- generic free code for array routines
 *
 *
 * DESCRIPTION:
 *	Release a variable length array, freeing each element.
 *
 *	Zeros space and elems and the array pointer
 *	after freeing the array.
 *
 * SEE ALSO:
 *	<fill in>
 *
 * RETURN VALUES:
 *	An errno value.
 *
 *--------------------------------------------------------------------------
 */
int
i405_free_array(there, here, what_d, elemsp, spacep, how, why)
	void *there;    /* IN */
	caddr_t *here;   /* INOUT (depends on the how argument) */
	struct xport_datatype *what_d;   /* IN: element type */
	int *elemsp;	/* INOUT: how much of array is filled. */
	int *spacep;	/* INOUT: space physically allocated for array */
	int how;        /* IN */
	int why;        /* IN */
{
#ifndef MIN
#define MIN(A,B) ((A) < (B) ? (A) : (B))
#endif
	int i, used_space = MIN(*elemsp, *spacep);
	caddr_t elhere;
	int elsize = what_d->size;
	int error;

	if (!WHY_FREE_P(why))
		panic("i405_free_array: bad why.");

	/*
	 * Now we xdr_free each element of array.
	 * We only do this for big arguments with their own sub-procedure.
	 * For little things without an elproc, we just skip all this stuff.
	 *
	 * NEEDSWORK: If we get an error freeing, we currently just
	 * go on.  (We must so that we can free all memory.)
	 * Is there a better way?
	 */
	if (elsize  >= sizeof(int))
		for (i = 0, elhere = *here; i < used_space; i++, elhere += elsize) {
			if (error = I405_XPORT_INTERPRET(there, elhere, what_d, how, why)) {
				printf ("i405_free_array: element free failed with error %d\n", error);
			};
		};

	/*
	 * Now free the array itself.
	 */
	if (*spacep)
		mem_free (*here, *spacep * elsize);
	*here = NULL;
	*elemsp = *spacep = 0;

	return 0;
}


/*
 *--------------------------------------------------------------------------
 *
 * i405_alloc_array -- generic alloc code for array routines
 *
 *
 * DESCRIPTION:
 *	Allocate a variable length array.
 *      Fill in space.
 *
 * SEE ALSO:
 *	<fill in>
 *
 * RETURN VALUES:
 *	An errno value.
 *
 *--------------------------------------------------------------------------
 */
int
i405_alloc_array(there, here, what_d, count, how, why)
	void *there;
	caddr_t *here;
	struct xport_datatype *what_d;   /* IN: element type */
	int count;	/* INOUT */
	int how;
	int why;
{
	int elsize = what_d->size;

	if (*here)
		panic ("i405_alloc_array: array already allocated.");

	if (elsize == 0)
		panic("i405_alloc_array: bogus elsize.");
	if (count == 0) {
		*here = NULL;
	} else {
		*here = mem_alloc (count * elsize);
		bzero(*here, count * elsize);
	};
	return 0;
}       




