
/*-
 *
 * null_subr.c
 *      vnode management code
 *
 *	$Id: null_subr.c,v 1.32 1994/10/03 10:49:23 johnh Exp $ (UCLA Ficus Project) $Date: 1994/10/03 10:49:23 $
 *
 * Copyright (c) 1992-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/user.h>
#include <sys/vnode.h>
#include <i405/i405.h>
#ifdef SVCM_TWO
#include <i405/SVCM/svcm.h>
#endif /* SVCM_TWO */
#include <sys/vfs.h>
#include <sys/vfs_stat.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/buf.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/pathname.h>
#include <vm/seg.h>
#include <sys/dnlc.h>
#include <sys/syslog.h>
#include <sys/debug.h>

#ifdef USE_NEWMAP
#include <sys/mman.h>
#include <vm/hat.h>
#include <vm/as.h>
#include <vm/seg_vn.h>
#endif USE_NEWMAP

#include <nullfs/null_node.h>
#include <nullfs/null_info.h>


#define NULL_HASH_BUCKETS 0x20
#define NULL_HASH_MASK (NULL_HASH_BUCKETS-1)
#define NULL_TO_HASH(LVP) ((((int)LVP)>>2) & NULL_HASH_MASK)
static struct null_node *null_hash[NULL_HASH_BUCKETS];

static struct null_node *null_node_base = NULL;   /* for kmem_fast_alloc */

static struct null_node *
null_alloc_nullnode()
{
	return (struct null_node *)kmem_fast_alloc((caddr_t *)&null_node_base,
					sizeof(struct null_node), 0x10);
}

void
null_dealloc_nullnode(np)
	struct null_node *np;
{
	kmem_fast_free((caddr_t *)&null_node_base, (caddr_t)np);
}

static void
add_nullnode_to_hash(np, lvp)
	struct null_node *np;
	struct vnode *lvp;
{
        int hash = NULL_TO_HASH(lvp);
	/*
	 * Put the null_node in the beginning of the bucket.
	 */
	np->null_next = null_hash[hash];
	null_hash[hash] = np;
}
	



/*
 *--------------------------------------------------------------------------
 *
 * null_find_nullnode -- look on the hash chains for a given node
 *
 * DESCRIPTION:
 *	Look on the hash chains for a given node.
 *	If we find it, hold and return it, otherwise return NULL.
 *
 * RETURN VALUES:
 *      Returns a (null-layer) vnode if one matches lvp, otherwise NULL.
 *
 *--------------------------------------------------------------------------
 */
static struct vnode *
null_find_nullnode (lvp, ninfo) 
	struct vnode *lvp;
	struct null_info *ninfo;
{
	struct null_node *np;
        int hash = NULL_TO_HASH(lvp);

        /*
	 * Check the hash chain.
	 */
#ifdef SVCM_TWO
 again:
#endif /* SVCM_TWO */
        np = null_hash[hash];
        while (np) {
                if ((NULLTOLOWERV(np) == lvp) &&
		    (NULLTOV(np)->v_vfsp == NULLINFOTOVFS(ninfo))) {
#ifdef SVCM_TWO
			/*
			 * Check to see if the node is going away.
			 * If so, wait for it to die and try again.
			 */
			if (np->null_flags & NULL_DIEING) {
				np->null_flags |= NULL_REBORN;
				sleep((caddr_t)NULLTOV(np), PINOD);
				goto again;
			};
#endif
			VN_HOLD(NULLTOV(np));
                        return NULLTOV(np);
                }
		else np = np->null_next;
        };
	/*
	 * Null_node is not present.
	 */
	return NULL;
}


/*
 *--------------------------------------------------------------------------
 *
 * null_make_nullnode -- build a new null node over a lower-level vnode
 *
 * DESCRIPTION:
 *	Null_make_nullnode is the primary entry point to the null-node
 *	management routines from the null_vnodeops.c.  We first look in
 *	the cache and re-use any cached vnode if available.
 *	If no such node exists, we build a new null node.
 *
 *	Note the invarient regarding reference counts and the null layer:
 *	each null layer node holds exactly one reference to the lower-layer
 *	node.  When re-using nodes we therefore release our (now found
 *	to be extra) hold on the lower-layer node.
 *
 * RETURN VALUES:
 *      A (null-layer) vnode.
 *
 *--------------------------------------------------------------------------
 */
struct vnode *
null_make_nullnode (lvp, ninfo, cred) 
	struct vnode *lvp;		/* IN: lower-layer vnode */
	struct null_info *ninfo;	/* IN: null-node vfs structure */
	struct ucred *cred;		/* IN */
{
	struct vnode *vp;
	struct null_node *np = NULL;
	int return_vp_after_cleanup = 0;
#ifdef SVCM_TWO
	USES_NVOP_CACHENAMEVP;
	struct vnode *cachevp;
	int error;
#endif /* SVCM_TWO */

	if (vp = null_find_nullnode(lvp, ninfo)) {
		/*
                 * Since the null-node corresponding to lvp already exists
                 * we should reduce the v_count of lvp which is bumped
                 * before null_make_nullnode is called.
		 */
                VN_RELE (lvp);
		return vp;
	}
	/*
	 * Null_node is not present, so create it.
	 */
	np = null_alloc_nullnode();
	vp = NULLTOV(np);
	bzero((char *)np, sizeof (struct null_node));
	np->null_lowervp = lvp;
	VTONULL(vp) = np;
	VN_INIT(vp, NULLINFOTOVFS(ninfo), lvp->v_type, lvp->v_rdev); 
#ifdef I_FWL
	vp->v_op = (struct vnodeops *) ninfo->null_vnodeop_p;
#else /* ! I_FWL */
	vp->v_op = (struct vnodeops *) null_vnodeop_p;
#endif /* I_FWL */
	/*
	 * All other vnode fields are zero.
	 */
	ninfo->info_refct++;
#ifdef SVCM_TWO
	/*
	 * Create an SVCM structure for our node.
	 */
	if (error = NVOP_CACHENAMEVP(vp, &cachevp, cred))
		goto out;
	if (error = svcm_name_create(&np->null_name, cachevp, NULL, cred))
		goto out;
	(void)svcm_unlock(np->null_name->lock);
#endif

	/*
	 * Race check.  We could sleep in memory allocation
	 * and someone else could come along and allocate our vnode
	 * for us.  If so, use theirs and free ours.
	 */
	if (vp = null_find_nullnode(lvp, ninfo)) {   /* race check */
		return_vp_after_cleanup = 1;
                VN_RELE (lvp);
		goto out;
	};
	/*
	 * Now that np is complete and there is no alias,
	 * add it to the hash list.
	 * This must be atomic with the null_find_nullnode call.
	 */
	add_nullnode_to_hash(np, lvp);

	return NULLTOV(np);

	/* Clean up and exit. */
 out:
	if (np) {
#ifdef SVCM_TWO
		if (np->null_name) {
			(void)svcm_lock(np->null_name->lock);
			svcm_name_destroy(np->null_name, NULLTOV(np), cred);
		};
#endif /* SVCM_TWO */
		null_dealloc_nullnode(np);
	};

	return return_vp_after_cleanup ? vp : NULL;
}


/*
 *--------------------------------------------------------------------------
 *
 * null_free_nullnode -- destroy a null node
 *
 * DESCRIPTION:
 *	The counterpart to null_make_nullnode, this routine
 *	is called when a nodes refcount drops to zero.
 *	We clean up after ourselves, releasing
 * 	the lowervp and throwing ourselves into freemem.
 *
 * RETURN VALUES:
 *      None.
 *
 *--------------------------------------------------------------------------
 */
void
null_free_nullnode (np)
	struct null_node *np;
{
	int hash = NULL_TO_HASH(NULLTOLOWERV(np));	
	struct null_node *ntablep;
	struct null_info *ninfo = VFSTONULLINFO(NULLTOV(np)->v_vfsp);

	if (ninfo->info_refct < 1) 
		panic ("null_free_nullnode: nullinfo refct is zero");

	/*
	 * First release the lowervp.
	 * (Notice that this must be *after* we compute the hash value.)
	 */
	VN_RELE(NULLTOLOWERV(np));
	NULLTOLOWERV(np) = NULL;   /* searchs on this node will now fail */

	ninfo->info_refct--;
        ntablep = null_hash[hash];
	if (ntablep == np) {
		null_hash[hash] = np->null_next;
	} else {
		while (ntablep) {
			if (ntablep->null_next == np) {
				ntablep->null_next = np->null_next;
				break;
			}
			else ntablep = ntablep->null_next;
		}
		if (ntablep == NULL)
			panic ("null_free_nullnode: null node not found in hash chain.");
	}

#ifdef SVCM_TWO
	/*
	 * Before freeing it, check for rebirth.
	 */
	if (np->null_flags & NULL_REBORN) {
		wakeup((caddr_t) NULLTOV(np));
	};
#endif

	/*
	 * Np can be freed.
	 */
	null_dealloc_nullnode(np);
}
		
	

#ifdef SVCM_TWO
int
null_flush_nullnodes(vfsp, cred)
	struct vfs *vfsp;
	struct ucred *cred;
{
	int i;
	struct null_node *np;
	struct vnode *vp;
	int needs_retry;
	USES_NVOP_PUTPAGE;

	for (i = 0; i < NULL_HASH_BUCKETS; i++) {
	retry:
		for (np = null_hash[i]; np; np = np->null_next) {
			vp = NULLTOV(np);
			/*
			 * Skip null nodes with no active pages.
			 * Skip vnodes which are going away.
			 * NEEDWORK: also skip locked nodes
			 *    and svcm locked nodes.
			 */
			if (vp->v_pages == NULL ||
					vp->v_count < 1 ||
					(np->null_flags & NULL_DIEING) != 0)
				continue;
			VN_HOLD(vp);
			ASSERT(vp->v_count >= 1);
			(void) NVOP_PUTPAGE(vp, 0, 0, B_ASYNC, cred);
			/*
			 * Inactivating the node may change the hash
			 * chains.  If so, we redo our hash chain.
			 */
			needs_retry = (vp->v_count <= 1);
			VN_RELE(vp);
			if (needs_retry)
				goto retry;
		};
	};
	return 0;
}
#endif    /* SVCM_TWO */



