
/*
 * null_vfsops.c
 *
 *	$Id: null_vfsops.c,v 1.21 1995/01/30 16:25:21 johnh Exp $ (UCLA Ficus Project) $Date: 1995/01/30 16:25:21 $
 *
 * Copyright (c) 1992-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/user.h>
#include <sys/file.h>
#include <sys/vfs.h>
#include <sys/vnode.h>
#include <i405/i405.h>
#include <sys/pathname.h>
#include <sys/uio.h>

#include <nullfs/null_info.h>
#include <nullfs/null_node.h>

#ifdef OUTOFKERNEL
#include <i405/ook_compat.h>
#endif

#if 0
#define DODEBUG(A) A
#else
#define DODEBUG(A)
#endif



/*
 * nullfs vfs operations.
 */
static int null_mount();
static int null_unmount();
static int null_root();
static int null_statfs();
static int null_sync();
static int null_vget();
static int null_mountroot();
static int null_swapvp();

struct vfsops null_vfsops = {
	null_mount,
	null_unmount,
	null_root,
	null_statfs,
	null_sync,
	null_vget,
	null_mountroot,
	null_swapvp,
};



/*
 * Map stuff is so that we can return correct device numbers.
 */
#define MAPSIZE  256/NBBY
static char null_minmap[MAPSIZE]; /* Map for minor device allocation */

int null_fsindex = -1;   /* Null layer fsindex.  Computed on first call to null_get_fsindex. */


/*
 *--------------------------------------------------------------------------
 *
 * null_mount -- the null layer's VFS_MOUNT
 *
 * DESCRIPTION:
 *	The plan: build our vfs structure and hand-construct
 *	our root vp.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_mount(vfsp, path, data
#ifdef I_FWL
	, fwl_list
#endif /* I_FWL */
	)
	struct vfs *vfsp;   /* IN: the vfs we should fill in */
	char *path;	    /* IN: path is where the new layer should go */
	caddr_t data;       /* IN: null-layer specific data */
#ifdef I_FWL
	char **fwl_list;
#endif /* I_FWL */
{
	int error;
	struct vnode *srootvp = NULL;
	struct vnode *lowerrootvp;
	struct null_mountargs args;
	struct null_info *ninfo = NULL;

#ifdef I_FWL
	if (null_fsindex == -1) {
		null_fsindex = i405_getfsindex("null");
	};
#endif /* I_FWL */

	error = copyin(data, (caddr_t)&args, sizeof (args));
	if (error) {
		return (error);
	}

	/*
	 * Find lower-level root, and make vfs point to lower vfs.
	 */
#ifdef OUTOFKERNEL
	error = vn_open (args.source, UIO_OURSPACE, 0, 0, &lowerrootvp);
#else
	error = lookupname(args.source, UIOSEG_USER, FOLLOW_LINK,
	    (struct vnode **)0, &lowerrootvp);
#endif
	if (error)
		goto out;

	ninfo = (struct null_info *)kmem_zalloc(sizeof (*ninfo));
	ninfo->info_lowervfs = lowerrootvp->v_vfsp;
	ninfo->info_vfs = vfsp;
	ninfo->info_enable_dnlc = args.enable_dnlc;
	ninfo->info_refct = 0;
	
	ninfo->info_mntno = vfs_getnum(null_minmap, MAPSIZE);
	vfsp->vfs_fsid.val[0] = ninfo->info_mntno;
	vfsp->vfs_fsid.val[1] = i405_getfsindex("null");
	
	vfsp->vfs_flag |= (NULLINFOTOLOWERVFS(ninfo)->vfs_flag & INHERIT_VFS_FLAG);
	vfsp->vfs_bsize = NULLINFOTOLOWERVFS(ninfo)->vfs_bsize;
	VFSTONULLINFO(vfsp) = ninfo;

	DODEBUG(printf("Null mount: lowerrootvp %x, vfsp %x\n", lowerrootvp, lowerrootvp->v_vfsp));
	DODEBUG(printf ("Null mount: dnlc %d\n", args.enable_dnlc));

#ifdef I_FWL
	if (error = fwl_mount_fwls(fwl_list, null_vnodeopv_list, &ninfo->null_vnodeop_p))
		goto out;
#endif /* I_FWL */
	srootvp = null_make_nullnode(lowerrootvp, ninfo, u.u_cred);
	DODEBUG(printf("Nmount: srootvp %x\n",srootvp));
	if (srootvp->v_flag & VROOT) {
		error = EBUSY;
		goto out;
	};
	srootvp->v_flag |= VROOT;
	
	ninfo->info_rootvp = srootvp;
	return 0;

out:
	if (srootvp)
		VN_RELE(srootvp);
#ifdef I_FWL
	if (ninfo->null_vnodeop_p)
		fwl_rele_fwls(null_vnodeopv_list, &ninfo->null_vnodeop_p);
#endif /* I_FWL */
	if (ninfo)
		kmem_free((caddr_t)ninfo, sizeof (*ninfo));
	return (error);
}


/*
 *--------------------------------------------------------------------------
 *
 * null_unmount -- the null layer's VFS_UNMOUNT
 *
 * DESCRIPTION:
 *	Just clean up after ourselves.
 *	We're responsible for the refct on the root node.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_unmount(vfsp)
	struct vfs *vfsp;   /* IN: vfs to unmount */
{
	struct null_info *ninfo;

	ninfo = VFSTONULLINFO(vfsp);
#ifdef NULL_USE_DNLC
	/*
	 * Purge the cache.  No other file systems seem to do this,
	 * it seems to me like the cache could be holding references
	 * to our vnodes, causing a false busy reporting.  -JSH
	 *
	 * No one else does this because it's done in os/vfs.c.
	 * Of course, this VFS_SYNC is done there too.  Oh well,
	 * what's wrong with a little excess.  -JSH
	 */
	dnlc_purge();
#endif
	VFS_SYNC(vfsp);
	if (ninfo->info_refct != 1 || ninfo->info_rootvp->v_count != 1) {
		return (EBUSY);
	}
	VN_RELE(ninfo->info_rootvp);
	kmem_free((caddr_t)ninfo, (u_int)sizeof (*ninfo));
	return (0);
}


/*
 *--------------------------------------------------------------------------
 *
 * null_root -- the null layer's VFS_ROOT
 *
 * DESCRIPTION:
 *	Just return the (HELD) root vnode.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_root(vfsp, vpp)
	struct vfs *vfsp;      /* IN: relevant vfs */
	struct vnode **vpp;    /* OUT: returned HELD root vnode */
{
	*vpp = (struct vnode *)VFSTONULLINFO(vfsp)->info_rootvp;
	(*vpp)->v_count++;
	return (0);
}

/*
 *--------------------------------------------------------------------------
 *
 * null_statfs -- the null layer's VFS_STATFS
 *
 * DESCRIPTION:
 *	Just pass it down.
 *	NEEDSWORK: Should munge f_fsid.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_statfs(vfsp, sbp)
	register struct vfs *vfsp;
	struct statfs *sbp;
{
	return (VFS_STATFS(VFSTOLOWERVFS(vfsp), sbp));
}


int null_enable_sync = 1;

/*
 *--------------------------------------------------------------------------
 *
 * null_sync -- the null layer's VFS_SYNC
 *
 * DESCRIPTION:
 *	Just pass it down.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_sync(vfsp)
	struct vfs *vfsp;
{
	int error=0;
#ifdef SVCM_TWO
	static int in_progress = 0;

	if (in_progress == 0) {
		in_progress++;
		if (null_enable_sync)
			error = null_flush_nullnodes(vfsp, u.u_cred);
	};
	in_progress = 0;
#endif
        if (!error && vfsp != (struct vfs *)NULL)
                error = VFS_SYNC(VFSTOLOWERVFS(vfsp));
	return error;
}

/*
 *--------------------------------------------------------------------------
 *
 * null_vget -- the null layer's VFS_VGET
 *
 * DESCRIPTION:
 *	Do the vget on the lower layer and then build a null-node
 *	around the returned vp.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_vget(vfsp, vpp, fidp)
	struct vfs *vfsp;
	struct vnode **vpp;
	struct fid *fidp;
{
	int error;
	struct null_info *ninfo = VFSTONULLINFO(vfsp);
	struct vfs *lowervfsp = NULLINFOTOLOWERVFS(ninfo);
	
	if (error = VFS_VGET(lowervfsp, vpp, fidp))
		return error;
	*vpp = null_make_nullnode (*vpp, ninfo, u.u_cred);
	return (0);
}

static int
null_mountroot(vfsp,vpp,name)
	struct vfs *vfsp;
	struct vnode **vpp;
	char *name;
{
	return EINVAL;
}

static int
null_swapvp(vfsp,vpp,name)
	struct vfs *vfsp;
	struct vnode **vpp;
	char *name;
{
	return EINVAL;
}


