
/*-
 *
 * null_vnodeops.c
 *
 *	$Id: null_vnodeops.c,v 1.43 1995/05/08 17:28:31 johnh Exp $ (UCLA Ficus Project) $Date: 1995/05/08 17:28:31 $
 *
 * Copyright (c) 1992-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 ***********************************************************************
 *
 * A null layer with the new interface.
 * Initial work by Yuguang Wu, DNLC, tuning by John Heidemann.
 * Vnode management code re-writen by Ashvin Goel to be SunOS-license
 * free.  SVCM interface by John Heidemann.
 *   
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/user.h>
#include <ficus/cdefs.h>
#include <sys/vnode.h>
#include <i405/i405.h>
#ifdef SVCM_TWO
#include <i405/SVCM/svcm.h>
#endif /* SVCM_TWO */
#include <sys/vfs.h>
#include <sys/vfs_stat.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/buf.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/pathname.h>
#include <vm/seg.h>
#include <sys/dnlc.h>
#include <sys/syslog.h>
#include <sys/debug.h>

#ifdef SVCM_TWO
#include <sys/mman.h>
#include <vm/hat.h>
#include <vm/as.h>
#include <vm/seg.h>
#include <vm/seg_vn.h>
#endif SVCM_TWO

#include <nullfs/null_node.h>
#include <nullfs/null_info.h>


#if 0
#define DODEBUG(A) A
#else
#define DODEBUG(A)
#endif


#define NULL_SANITY_CHECKS
#ifdef NULL_SANITY_CHECKS
int null_sanity_checks = 1;
#endif /* NULL_SANITY_CHECKS */


/*-
 *
 * The basic null layer requires only that bypass and inactive be
 * implemented.  Inactive is needed because it frees nullnodes
 * that are no longer needed.
 *
 * To this list, a few routines are added:
 *   - Open is needed because it has the funny first argument
 *     (an obnoxious struct vnode **, not just a struct vnode *).
 *   - Getattr is needed so that we can assign a unique fsid
 *     to nullnode's attrs.  This is not really essential,
 *     but if it's not done then pwd can break if the null layer
 *     and it's source are mounted in the same directory.
 *
 * Currently, directory name lookup caching is mostly implemented.
 * Lookup adds things to the cache.  Rename, remove, and rmdir
 * all purge the cache appropriately.  As a NEEDSWORK, it might
 * be nice to have name lookup caching added to mkdir and create,
 * but the payoff didn't seem to be worth the effort.
 *
 *
 * While we're on the topic of the dnlc, I should mention why
 * it's such a good idea.  The UFS and NFS (I presume) have fancy
 * xnode caching code.  The null layer doesn't have this, when
 * a nullnode goes inactive, we do a kmem_free on the darn thing.
 * This is bad if you are parsing many long but similar pathnames,
 * because the upper levels of the tree will have their nullnodes
 * repeatedly built and freed, wasting a lot of time.
 *
 * The dnlc comes in because it holds a reference count on nullnodes.
 * Because of this count, they're not freed when you're done
 * parsing that element of the path, and they are presumably
 * still around in the cache when we need them again.
 *
 * I guess it's not a bad sort of a hack.  It seems better than
 * making grossly complicated nullnode management code, when the
 * null layer should be as simple as possible.
 *
 * If you're mounting several null layers on top of each other
 * (say, for benchmarking), using caching on all layers is bad
 * because you crowd the cache with a bunch of nullnodes which
 * will already have their refcounts bumped.  Only do caching
 * on the top-most layer.
 *    -JSH, 21-Nov-90
 *
 * [This point about caching is true in general:  you usually 
 * only want the top layer to actively cache things.  -JSH, 8-Mar-93]
 *
 * Null layer performance isn't as good as might be expected.
 * The MAB costs .4% per layer, but recursive copy is 3% and remove
 * is 6%.  To improve this, dnlc'ing is added to mkdir and create,
 * and null_make_nullnode is changed to use kmem_fast_alloc.
 *    -JSH, 17-Dec-90
 *
 * These numbers have improved since.  MAB and other elapsed time benchmarks
 * run about 0.3% per layer.  Recursive copy is about 1.7% overhead,
 * and recursive remove about 2.3% or so.
 * See the Master's Thesis for details.
 *    -JSH, 8-Feb-91
 *
 *
 * With DNLC cache coherent lower layers the hand-coded routines
 * to manage our cache are no longer needed.
 * #undef NON_SVCM_LOWER_LAYER
 * if you can guarantee that your lower layer will abide by the SVCM. 
 *
 */



int null_bug_bypass = 0;    /* a good variable for a debugger to change */




/*-
 *--------------------------------------------------------------------------
 *
 * null_bypass -- a prototypical bypass routine
 *
 *
 * DESCRIPTION:
 *	The basic plan:
 *		map the vnodes, call the op and then map back.
 * 	See the comments for details.
 *
 *	This version was revised 10-Apr-92 for new vnodeop_desc structure.
 *	   This version has been optimized for speed, throwing away some
 *	safety checks.  It should still always work, but it's not as
 *	robust to programmer errors.
 *	   Define SAFETY to include some error checking code.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_bypass(ap)
	struct nvop_generic_args *ap;
{
	int error;
	struct vnode *old_vps[VDESC_MAX_VPS];
	struct vnode **vps_p[VDESC_MAX_VPS];
	struct vnode ***vppp;
	struct vnodeop_desc *descp = ap->a_desc;
	int flags, i;

	if (null_bug_bypass)
		printf ("null_bypass: %s\n", descp->vdesc_name);

#ifdef SAFETY
	/*
	 * We want to make sure we have at least one vp.
	 * This should throw out calls by Sun's open.
	 *
	 */
	if (descp->vdesc_vp_offsets == NULL ||
	    descp->vdesc_vp_offsets[0] == VDESC_NO_OFFSET)
		panic ("null_bypass: no vp's in map.\n");
#endif

	/*
	 * Map the vnodes going in.
	 * Later, we'll invoke the operation based on
	 * the mapped, first vnode's operation vector.
	 */
	flags = descp->vdesc_flags;
	for (i = 0; i < VDESC_MAX_VPS; flags >>= 1, i++) {
		if (descp->vdesc_vp_offsets[i] == VDESC_NO_OFFSET)
			break;   /* bail out at end of list */
		if (flags & 1)   /* skip vps that aren't to be mapped */
			continue;
		vps_p[i] = VOPARG_OFFSETTO(struct vnode**,descp->vdesc_vp_offsets[i],ap);
		old_vps[i] = *(vps_p[i]);
		*(vps_p[i]) = NULLTOLOWERV(VTONULL(*(vps_p[i])));
	};

	NULL_VERIFY_BOTH_PAGES(old_vps[0]);

	/*
	 * We assume that any vpp's are only OUT values.
	 * If they're INOUT values (as with that pesky vop_open)
	 * we'd need to do a little more work here.
	 */

	/*
	 * We assume that vp1p is the vnode of choice
	 * (with the desired operation vector),
	 * and that it's been correctly mapped above.
	 *
	 * Call the operation on the lower layer
	 * with the modified argument structure.
	 */
	error = VCALL(*(vps_p[0]), descp->vdesc_offset, ap);

	/*
	 * Maintain the illusion of call-by-value
	 * by restoring vnodes in the argument structure
	 * to their original value.
	 */
	flags = descp->vdesc_flags;
	for (i = 0; i < VDESC_MAX_VPS; flags >>= 1, i++) {
		if (descp->vdesc_vp_offsets[i]==VDESC_NO_OFFSET)
			break;   /* bail out at end of list */
		if (flags&1)   /* skip vps that aren't to be mapped */
			continue;
		*(vps_p[i]) = old_vps[i];
	};

	/*
	 * Map the possible out-going vpp.
	 */
	if (descp->vdesc_vpp_offset != VDESC_NO_OFFSET &&
	    !(descp->vdesc_flags & VDESC_NOMAP_VPP) &&
	    !error) {
		struct ucred **credpp = VOPARG_OFFSETTO(struct ucred**, 
					 descp->vdesc_cred_offset, ap);
		vppp = VOPARG_OFFSETTO(struct vnode***,
				 descp->vdesc_vpp_offset, ap);
		**vppp = null_make_nullnode(**vppp,
					    VFSTONULLINFO(old_vps[0]->v_vfsp),
					    *credpp);
		NULL_VERIFY_BOTH_PAGES(**vppp);
	};

	NULL_VERIFY_BOTH_PAGES(old_vps[0]);

	return (error);
}



/*-
 *--------------------------------------------------------------------------
 *
 * null_getattr -- null layer getattr routine which maps va_fsid
 *
 *
 * DESCRIPTION:
 *	Getattr is required because we have to change the va_fsid field.
 *	We must do this to get proper pwd(1) behavior when the null layer
 *	and its lower layer are in the same directory.
 *
 *	SVCM:  There's no need for us to to cache coherence here.
 *	The lower layer does it for us.  Of course, once the attributes
 *	are returned they're no longer (necessarily) valid.  If the higher
 * 	level client wants guaranteed coherence, it should have locked
 *	the cache to prevent other writers.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_getattr(ap)
	struct nvop_getattr_args *ap;
{
	int error;
	if (error = null_bypass(ap))
		return error;
	/* Requires that arguments be restored. */
 	ap->a_vap->va_fsid = i405_generate_unique_fsid(ap->a_vp->v_vfsp,
				 VFSTONULLINFO(ap->a_vp->v_vfsp)->info_mntno);
	return 0;
}


/*-
 *--------------------------------------------------------------------------
 *
 * null_getattr -- null layer open routine to cope with open's vpp
 *
 * DESCRIPTION:
 *	Open is here because it's irregular, like all those French verbs.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_open(ap)
	struct nvop_open_args *ap;
{
	int error;
        struct vnode **old_vpp;
        struct vnode *old_lower_vp, *new_lower_vp;

        old_vpp = ap->a_vpp;
        new_lower_vp = old_lower_vp = NULLTOLOWERV(VTONULL(*ap->a_vpp));
        ap->a_vpp = &new_lower_vp;
        error = VCALL(*ap->a_vpp, VOFFSET(nvop_open), ap);
        if (!error) {
                if (new_lower_vp != old_lower_vp) {
			VN_RELE(*old_vpp);
                        *old_vpp = null_make_nullnode(new_lower_vp,
					VFSTONULLINFO((*old_vpp)->v_vfsp),
						      ap->a_cred);
		};
        };
        ap->a_vpp = old_vpp;
	NULL_VERIFY_BOTH_PAGES(*ap->a_vpp);
        return error;
}


/*-
 *--------------------------------------------------------------------------
 *
 * null_getattr -- null layer inactive
 *
 * DESCRIPTION:
 *	Inactive frees our layer's node when the refct goes to zero.
 *	Ordering is very important here:  we have to deal with
 *	node destroy/create race conditions.  The general goal is to
 *	make sure that once something starts going away it really goes
 *	away even if someone else is trying to re-create it.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_inactive(ap)
	struct nvop_inactive_args *ap;
{
	struct vnode *vp = ap->a_vp;
	struct null_node *np = VTONULL(vp);
#ifdef SVCM_TWO
	int error;
#endif /* SVCM_TWO */

	DODEBUG(printf("null_inactive(%x,%x)\n",
		       (u_int)ap->a_vp, (u_int)ap->a_cred));
	if (vp->v_count != 0)
		panic("null_inactive: v_count != 0");

	/*
	 * Mark the node as in its last death throws.
	 * It cannot be re-used once this flag is set
	 * since it may be partially gone.
	 */
	np->null_flags |= NULL_DIEING;

	NULL_VERIFY_BOTH_PAGES(vp);

#ifdef SVCM_TWO
	/*
	 * Unname purges any cached pages
	 * and destroys all cache mapping information.
	 */
	if (error = svcm_lock(np->null_name->lock))
		panic ("null_inactive: failed to get lock.");
	svcm_name_rele(np->null_name, vp, ap->a_cred);
	if (vp->v_pages != NULL)
		panic("null_inactive: still have pages after svcm_name_rele");
#endif /* SVCM_TWO */

#ifdef JOHNH_DATA_PAGE_DEBUGGING
	if (null_special_null_vp == vp) {
		null_special_null_vp = NULL;
		/* null_special_ufs_vp = NULL; */
	};
#endif /* JOHNH_DATA_PAGE_DEBUGGING */

	NULL_VERIFY_PAGES(vp);

	/*
	 * null_free_nullnode will release the lower level node
	 * and then ours.
	 */
	null_free_nullnode(VTONULL(vp));
	return 0;
}




/**********************************************************************
 *
 *  DNLC CODE.
 *
 */


#if defined(NULL_USE_DNLC)
/*
 * These routines control handle directory name lookup caching
 * in the null layer.
 */


/*
 * This #define bogosity here is to provide code that works
 * with both SVCM and non-SVCM kernels without duplicating *everything*.
 * To read the code, pretend everyone is SVCM and then be confident
 * that the right things go away for non-SVCM people.
 */
#ifdef SVCM_TWO
#define SVCM_ONLY(X) X
#else /* ! SVCM_TWO */
#define SVCM_ONLY(X) 
#define cache_name NULL
#define tcache_name NULL
#define NON_SVCM_LOWER_LAYER
#define svcm_dnlc_lookup(NAME,DVP,NM,CRED)	dnlc_lookup(DVP,NM,CRED)
#define svcm_dnlc_purge()		dnlc_purge()
#define svcm_dnlc_enter(NAME, DVP, NM, VP, CRED) dnlc_enter(DVP, NM, VP, CRED)
#define svcm_dnlc_remove(NAME, DVP, NM, CRED) dnlc_remove(DVP, NM)
#endif /* SVCM_TWO */

#define DOING_DNLC(VP) (VFSTONULLINFO((VP)->v_vfsp)->info_enable_dnlc)

static int
null_lookup(ap)
	struct nvop_lookup_args *ap;
{
	int error;
	struct vnode *vp;
	struct vnode *dvp = ap->a_vp;
	SVCM_ONLY(struct svcm_name *cache_name = VTONULL(dvp)->null_name);

	/*
	 * First try the cache.
	 */
	if (vp = svcm_dnlc_lookup(cache_name, dvp, ap->a_nm, ap->a_cred)) {
		VN_HOLD(vp);
		*ap->a_vpp = vp;
		return 0;
	};

       	/*
	 * call bypass to do the command.
	 */
	if (error = null_bypass(ap))
		return error;
	/*
	 * No error returned, so cache the lookup.
	 */
	if (DOING_DNLC(ap->a_vp))
		svcm_dnlc_enter(cache_name, dvp, ap->a_nm, *ap->a_vpp, ap->a_cred);
	return 0;
}

#ifdef NON_SVCM_LOWER_LAYER
static int
null_rename(ap)
	struct nvop_rename_args *ap;
{
	int error;
	struct vnode *vp = ap->a_vp;
	SVCM_ONLY(struct svcm_name *cache_name = VTONULL(vp)->null_name);
	SVCM_ONLY(struct svcm_name *tcache_name = VTONULL(ap->a_tdvp)->null_name);

	/*
	 * Following NFS' example, we're conserative about purging
	 * the cache.  We purge it before the lower call, not after.
	 */
	if (DOING_DNLC(vp)) {
		if (error = svcm_dnlc_remove(cache_name,
					     vp,
					     ap->a_nm, ap->a_cred))
			return error;
		if (error = svcm_dnlc_remove(tcache_name,
					     ap->a_tdvp,
					     ap->a_tnm, ap->a_cred))
			return error;
	};

	/*
	 * Do the job.
	 */
	return null_bypass(ap);
}
#endif /* NON_SVCM_LOWER_LAYER */


#ifdef NON_SVCM_LOWER_LAYER
static int
null_remove(ap)
	struct nvop_remove_args *ap;
{
	int error;
	struct vnode *vp = ap->a_vp;
	SVCM_ONLY(struct svcm_name *cache_name = VTONULL(vp)->null_name);

	if (error = svcm_dnlc_remove(cache_name,
				     vp,
				     ap->a_nm, ap->a_cred))
		return error;
	return null_bypass(ap);
}
#endif /* NON_SVCM_LOWER_LAYER */


#ifdef NON_SVCM_LOWER_LAYER
static int
null_rmdir(ap)
	struct nvop_rmdir_args *ap;
{
	int error;
        struct vnode *vp = ap->a_vp;
        SVCM_ONLY(struct svcm_name *cache_name = VTONULL(vp)->null_name);

	if (error = svcm_dnlc_remove(cache_name,
				     vp,
				     ap->a_nm, ap->a_cred))
		return error;
	return null_bypass(ap);
}
#endif /* NON_SVCM_LOWER_LAYER */

#endif NULL_USE_DNLC



/**********************************************************************
 *
 *  SVCM CODE.
 *
 */

#ifdef SVCM_TWO
/*
 * SVCM_TWO code is experimental.
 */



/*-
 *--------------------------------------------------------------------------
 *
 * null_getpage -- null layer's vop_getpage, map in a page (SVCM only)
 *
 * DESCRIPTION:
 *	First aquire page ownership, then map it in directly with
 *	VOP_NEWGETPAGE.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_getpage (ap)
	struct nvop_getpage_args *ap;
{
	USES_NVOP_STACKGETPAGE;
	int error = 0;
	struct vnode *vp = ap->a_vp;
	struct null_node *np = VTONULL(vp);

	DODEBUG(if (null_bug_bypass)
		printf ("null_getpage (%x,%x,%x,%x,%x,%x,%x,%x,%d,%x)\n",
			vp,
			ap->a_off, ap->a_len,
			ap->a_protp,
			ap->a_pl, ap->a_plsz,
			ap->a_seg, ap->a_addr,
			ap->a_rw, ap->a_cred));

	if (error = svcm_lock(VTONULL(vp)->null_name->lock))
		return error;

	NULL_VERIFY_BOTH_PAGES(vp);

#ifdef NULL_SANITY_CHECKS
	if (!svcm_islocked(np->null_name->lock))
		panic("null_getpage: unlocked lock.");
#endif /* NULL_SANITY_CHECKS */
	DODEBUG(pvn_dump_pages(vp));
	error = NVOP_STACKGETPAGE(NULLTOLOWERV(np),
			       VTONULL(vp)->null_name, vp, 
			       ap->a_off, ap->a_len,
			       ap->a_protp,
			       ap->a_pl, ap->a_plsz,
			       ap->a_seg, ap->a_addr,
			       ap->a_rw, ap->a_cred);
#ifdef NULL_SANITY_CHECKS
	if (!error && null_sanity_checks) {
		pvn_verify_page_owner(vp);
		i405_vm_verify_pagelist_owner(ap->a_pl, ap->a_plsz, vp);
	};
#endif /* NULL_SANITY_CHECKS */

	NULL_CHECK_NULL_UFS_VP(ap->a_vp);
	NULL_VERIFY_BOTH_PAGES(vp);

	DODEBUG(pvn_dump_pages(vp));
	(void)svcm_unlock(np->null_name->lock);
	return error;
}


/*-
 *--------------------------------------------------------------------------
 *
 * null_putpage -- null layer's vop_putpage, map in a page (SVCM only)
 *
 * DESCRIPTION:
 *	First aquire page ownership, then map it in directly with
 *	VOP_NEWPUTPAGE.
 *
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_putpage (ap)
	struct nvop_putpage_args *ap;
{
	USES_NVOP_STACKPUTPAGE;
	int error;
	struct vnode *vp = ap->a_vp;
	struct null_node *np = VTONULL(vp);

	DODEBUG(if (null_bug_bypass)
		printf ("null_putpage (%x,%x,%x,%x,%x)\n",
			vp,
			ap->a_off, ap->a_len,
			ap->a_flags,
			ap->a_cred));

	if (!np->null_name)
		panic ("null_putpage: no name allocated.");
	if (error = svcm_lock(np->null_name->lock))
		return error;
	NULL_VERIFY_BOTH_PAGES(vp);

	/*
	 * Normally getpage will fill in svcm_name,
	 * but if the user's writing a new file,
	 * pvn_rdwr may just alloc its own pages, and
	 * so no name may be aquired.
	 *    HMMM...for svcm ownership to work,
	 * we assume getpage will aquire ownership.
	 * In the just-discussed case, it won't.
	 *    The solution is to force pvn_rdwr to pre-allocate pages,
	 * going through getpage and the svcm.  This is also
	 * important because the ufs needs to allocate backing store
	 * at that time.
	 *
	 * The page-out daemon also calls putpage,
	 * perhaps even before getpage has done its thing,
	 * so there are times when we mustwe _have_ to do this here.
	 *
	 * But all real pages _must_ be allocated through
	 * getpage.  Any other approach is an error.
	 * To detect this bug, the newputpage should
	 * check (if we want to be paranoid).
	 */
#if 0
	if (!vp->v_svcm_name)
		if (error = null_name_vp(vp, ap->a_cred))
			return errror;
#else
	/*
	 * Panic is above.
	 */
#endif
	DODEBUG(pvn_dump_pages(vp));
	error = NVOP_STACKPUTPAGE(NULLTOLOWERV(np),
			       np->null_name, vp,
			       ap->a_off, ap->a_len,
			       ap->a_flags,
			       ap->a_cred);
	DODEBUG(pvn_dump_pages(vp));

	NULL_CHECK_NULL_UFS_VP(ap->a_vp);
	NULL_VERIFY_BOTH_PAGES(vp);

	(void)svcm_unlock(np->null_name->lock);
	return error;
}


/*-
 *--------------------------------------------------------------------------
 *
 * null_rdwr -- the null layer's vop_rdwr (SVCM only)
 *
 * DESCRIPTION:
 *	Rdwr must be intercepted because we must force it through
 *	_our_ page cache, otherwise we access our lower-level's page
 *	cache and end up aliasing our own pages.
 *
 *	We just get the svcm locks and call the lower-layer's stackrdwr
 *	routine.
 *
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_rdwr (ap)
	struct nvop_rdwr_args *ap;
{
	USES_NVOP_STACKRDWR;
	int error;
	struct vnode *vp = ap->a_vp;
	struct null_node *np = VTONULL(vp);

	DODEBUG(if (null_bug_bypass)
		printf ("null_rdwr()\n"));

	if (vp->v_type != VREG)
		return null_bypass(ap);

	ASSERT(np->null_name);

	/*
	 * Don't do special file types (like directories)
	 * since they're not cache coherent.
	 */
	if (error = svcm_lock(np->null_name->lock))
		return error;

	NULL_VERIFY_BOTH_PAGES(vp);

	error = NVOP_STACKRDWR(NULLTOLOWERV(np),
			       np->null_name, vp,
			       ap->a_uiop, ap->a_rw,
			       ap->a_ioflag, ap->a_cred);

	NULL_CHECK_NULL_UFS_VP(vp);
	NULL_VERIFY_BOTH_PAGES(vp);
	
	(void)svcm_unlock(np->null_name->lock);

	return error;
}



/*-
 *--------------------------------------------------------------------------
 *
 * null_fsync -- null layer's vop_fsync (SVCM only)
 *
 * DESCRIPTION:
 *	Fsync is intercepted because it must work on our level.
 *	This is another good candidate to turn into a vn_fsync.
 *
 * RETURN VALUES:
 *      Returns an errno value.
 *
 *--------------------------------------------------------------------------
 */
static int
null_fsync(ap)
	struct nvop_fsync_args *ap;
{
	USES_NVOP_PUTPAGE;
	DODEBUG(printf("null_fsync on %x.\n", ap->a_vp));
	return NVOP_PUTPAGE(ap->a_vp, 0, 0, 0, ap->a_cred);
}


#ifdef JOHNH_DATA_PAGE_DEBUGGING
static int
null_map (ap)
	struct nvop_map_args *ap;
{
	int error = i405_vn_map(ap);
	null_special_null_vp = ap->a_vp;
	null_special_ufs_vp = VTONULL(ap->a_vp)->null_name->vp;
	return error;
}
#endif /* JOHNH_DATA_PAGE_DEBUGGING */


#ifdef I_FWL
static int
null_get_svcm_name(ap)
	struct nvop_get_svcm_name_args *ap;
{
	*ap->a_name_p = VTONULL(ap->a_vp)->null_name;
	return 0;
}


static int
null_get_fsindex(ap)
	struct nvop_get_fsindex_args *ap;
{
	*ap->a_fsindex = null_fsindex;
	return 0;
}
#endif /* I_FWL */


#endif SVCM_TWO



/**********************************************************************
 *
 *  OPS VECTOR.
 *
 */

int (**null_vnodeop_p)();

struct vnodeopv_entry_desc null_vnodeop_entries[] = {
	{ &nvop_default_desc, null_bypass },
	{ &nvop_inactive_desc, null_inactive },
	{ &nvop_open_desc, null_open },
	{ &nvop_getattr_desc, null_getattr },
#if defined(NULL_USE_DNLC)
	{ &nvop_lookup_desc, null_lookup },
#ifdef NON_SVCM_LOWER_LAYER
	{ &nvop_rename_desc, null_rename },
	{ &nvop_remove_desc, null_remove },
	{ &nvop_rmdir_desc, null_rmdir },
#endif /* NON_SVCM_LOWER_LAYER */
#endif NULL_USE_DNLC
#ifdef SVCM_TWO
	{ &nvop_rdwr_desc, null_rdwr },
#ifdef JOHNH_DATA_PAGE_DEBUGGING
	{ &nvop_map_desc, null_map },
#else /* ! JOHNH_DATA_PAGE_DEBUGGING */
	{ &nvop_map_desc, i405_vn_map },
#endif /* JOHNH_DATA_PAGE_DEBUGGING */
	{ &nvop_getpage_desc, null_getpage },
	{ &nvop_putpage_desc, null_putpage },
	{ &nvop_fsync_desc, i405_vn_fsync },
	{ &nvop_cache_callback_desc, i405_vn_cache_callback },
#ifdef I_FWL
	{ &nvop_get_svcm_name_desc, null_get_svcm_name },
	{ &nvop_get_fsindex_desc, null_get_fsindex },
#endif /* I_FWL */
#endif /* SVCM_TWO */
	{ NULL, NULL }
};
struct vnodeopv_desc null_vnodeop_opv_desc =
	{ &null_vnodeop_p, null_vnodeop_entries };

#ifdef I_FWL
struct vnodeopv_desc *null_vnodeopv_list[] = {
	&null_vnodeop_opv_desc,
	NULL
};
#endif /* I_FWL */
